#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Core stream definitions. """

from amaranth          import Elaboratable, Signal, Module
from amaranth.hdl.rec  import Record, DIR_FANIN, DIR_FANOUT
from amaranth.hdl.xfrm import DomainRenamer

from ..stream          import StreamInterface
from ..stream.arbiter  import StreamArbiter


class USBInStreamInterface(StreamInterface):
    """ Variant of LUNA's StreamInterface optimized for USB IN transmission.

    This stream interface is nearly identical to StreamInterface, with the following
    restriction: the `valid` signal _must_ be held high for every packet between `first`
    and `last`, inclusively.

    This means that the relevant interface can easily be translated to the UTMI transmit
    signals, with the following mappings:

        Stream  | UTMI
        --------|-----------
        valid   | tx_valid
        payload | tx_data
        ready   | tx_ready
    """

    def bridge_to(self, utmi_tx):
        """ Generates a list of connections that connect this stream to the provided UTMITransmitInterface. """

        return [
            utmi_tx.valid  .eq(self.valid),
            utmi_tx.data   .eq(self.payload),

            self.ready     .eq(utmi_tx.ready)
        ]



class USBOutStreamInterface(Record):
    """ Variant of LUNA's StreamInterface optimized for USB OUT receipt.

    This is a heavily simplified version of our StreamInterface, which omits the 'first',
    'last', and 'ready' signals. Instead, the streamer indicates when data is valid using
    the 'next' signal; and the receiver must keep time.

    This is selected so the relevant interface can easily be translated to the UTMI receive
    signals, with the following mappings:

        UTMI      | Stream
        --------- |-----------
        rx_active | valid
        rx_data   | payload
        rx_valid  | next

    """

    def __init__(self, payload_width=8):
        """
        Parameter:
            payload_width -- The width of the payload packets.
        """
        super().__init__([
            ('valid',    1,             DIR_FANOUT),
            ('next',     1,             DIR_FANOUT),

            ('payload',  payload_width, DIR_FANOUT),
        ])


    def bridge_to(self, utmi_rx):
        """ Generates a list of connections that connect this stream to the provided UTMIReceiveInterface. """

        return [
            self.valid     .eq(utmi_rx.rx_active),
            self.next      .eq(utmi_rx.rx_valid),
            self.data      .eq(utmi_rx.payload)
        ]


    def stream_eq(self, other):
        """ Generates a list of connections that connect this stream to the provided UTMIReceiveInterface. """
        return self.connect(other)




class USBOutStreamBoundaryDetector(Elaboratable):
    """ Gateware that detects USBOutStream packet boundaries, and generates First and Last signals.

    As UTMI/ULPI do not denote the last byte of a packet; this module injects two bytes of delay in
    order to correctly identify the last bytes.

    Attributes
    ----------
    unprocessed_stream: USBOutStreamInterface, input stream
        The stream to work with; will be processed and then output on :attr:``processed_stream``.
    processed_stream: USBOutStreamInterface, output stream
        The stream produced by this module. This stream is two bytes delayed from :attr:``unprocessed_stream``;
        and in-phase with the :attr::``first`` and :attr::``last`` signals.

    complete_in: Signal(), input, optional
        Input that accepts an RxComplete signal. If provided; a delayed version will be produced on
        :attr:``complete_out`` after a :attr:``processed_stream`` packet terminates.
    invalid_in: Signal(), input, optional
        Input that accepts an RxInvalid signal. If provided; a delayed version will be produced on
        :attr:``complete_out`` after a :attr:``processed_stream`` packet terminates.


    complete_out: Signal(), output
        If :attr:``complete_in`` is provided; this signal provides a delayed version of that signal
        timed so it is strobed after :attr:``processed_stream`` packets complete.
    invalid_out: Signal(), output
        If :attr:``invalid_out`` is provided; this signal provides a delayed version of that signal
        timed so it is strobed after :attr:``processed_stream`` packets complete.

    first: Signal(), output
        Indicates that the byte present on :attr:``processed_stream`` is the first byte of a packet.
    last: Signal(), output
        Indicates that the byte present on :attr:``processed_stream`` is the last byte of a packet.

    Parameters
    ----------
    domain: str
        The name of the domain the stream belongs to; defaults to "usb".
    """

    def __init__(self, domain="usb"):

        self._domain = domain

        #
        # I/O port
        #
        self.unprocessed_stream = USBOutStreamInterface()
        self.processed_stream   = USBOutStreamInterface()

        self.complete_in        = Signal()
        self.invalid_in         = Signal()

        self.complete_out       = Signal()
        self.invalid_out        = Signal()

        self.first              = Signal()
        self.last               = Signal()


    def elaborate(self, platform):
        m = Module()

        in_stream  = self.unprocessed_stream
        out_stream = self.processed_stream

        # We'll buffer a single byte of the stream, so we can always be one byte ahead.
        buffered_byte = Signal(8)
        is_first_byte = Signal()

        buffered_complete = Signal()
        buffered_invalid  = Signal()

        with m.FSM(domain='usb'):

            # WAIT_FOR_FIRST_BYTE -- we're not actively receiving data, yet. Wait for the
            # first byte of a new packet.
            with m.State('WAIT_FOR_FIRST_BYTE'):
                m.d.usb += out_stream.valid.eq(0)

                m.d.usb += [
                    # We have no data to output, so this can't be our first or last bytes...
                    self.first       .eq(0),
                    self.last        .eq(0),
                    out_stream.next  .eq(0),

                    # ... and we can't have gotten a complete or invalid strobe that matters to us.
                    buffered_complete     .eq(0),
                    buffered_invalid      .eq(0),
                    self.complete_out     .eq(0),
                    self.invalid_out      .eq(0),
                ]

                # Once we've received our first byte, buffer it, and mark it as our first byte.
                with m.If(in_stream.valid & in_stream.next):
                    m.d.usb += [
                        buffered_byte.eq(in_stream.payload),
                        is_first_byte.eq(1)
                    ]
                    m.next = 'RECEIVE_AND_TRANSMIT'

            # RECEIVE_AND_TRANSMIT -- receive incoming bytes, and transmit our buffered bytes.
            # We'll transmit one byte per byte received; ensuring we always retain a single byte --
            # our last byte.
            with m.State('RECEIVE_AND_TRANSMIT'):
                m.d.usb += [
                    out_stream.valid  .eq(1),
                    out_stream.next   .eq(0)
                ]

                # Buffer any complete/invalid signals we get while receiving, so we don't output
                # them before we finish outputting our processed stream.
                m.d.usb += [
                    buffered_complete  .eq(buffered_complete | self.complete_in),
                    buffered_invalid   .eq(buffered_invalid  | self.invalid_in)
                ]

                # If we get a new byte, emit our buffered byte, and store the incoming byte.
                with m.If(in_stream.valid & in_stream.next):
                    m.d.usb += [
                        # Output our buffered byte...
                        out_stream.payload  .eq(buffered_byte),
                        out_stream.next     .eq(1),

                        # indicate whether our current byte was the first byte captured...
                        self.first          .eq(is_first_byte),

                        # ... and store the new, incoming byte.
                        buffered_byte       .eq(in_stream.payload),
                        is_first_byte       .eq(0)
                    ]

                # Once we no longer have an active packet, transmit our _last_ byte,
                # and move back to waiting for an active packet.
                with m.If(~in_stream.valid):
                    m.d.usb += [

                        # Output our buffered byte...
                        out_stream.payload  .eq(buffered_byte),
                        out_stream.next     .eq(1),
                        self.first          .eq(is_first_byte),

                        # ... and indicate that it's the last byte in our stream.
                        self.last           .eq(1)
                    ]
                    m.next = 'OUTPUT_STROBES'

            with m.State('OUTPUT_STROBES'):
                m.d.usb += [
                    # We've just finished transmitting our processed stream; so clear our data strobes...
                    self.first        .eq(0),
                    self.last         .eq(0),
                    out_stream.next   .eq(0),

                    # ... and output our buffered complete/invalid strobes.
                    self.complete_out .eq(buffered_complete),
                    self.invalid_out  .eq(buffered_invalid)
                ]
                m.next = 'WAIT_FOR_FIRST_BYTE'


        if self._domain != "usb":
            m = DomainRenamer({"usb": self._domain})(m)

        return m


class USBRawSuperSpeedStream(StreamInterface):
    """ Variant of LUNA's StreamInterface optimized for carrying raw USB3 data.

    Low-level USB3 data-streams consist of both data bytes ("data") and control flags,
    which differentiate standard data bytes from data bytes used for control.

    This variant comes implicitly with the relevant control flags; and is sized to allow
    gearing that makes USB3's high-speed signals manageable.

    Parameters
    ----------
    payload_words: int
        The number of payload words (1 byte data, 1 bit control) to include in the current stream.
    """

    def __init__(self, payload_words=4):
        super().__init__(payload_width=8 * payload_words, extra_fields=[('ctrl', payload_words)])


    def stream_eq(self, interface, *, endian_swap=False, omit=None, **kwargs):
        """ Extend the global ``stream_eq`` operator to swap endianness. """

        # If we're not performing an endian swap, delegate directly to our parent.
        if endian_swap == False:
            return super().stream_eq(interface, omit=omit, **kwargs)

        # Otherwise, perform our full endian swap.

        if omit is None:
            omit = []

        # Add ``data`` and ``ctrl`` to the list of fields to omit, as we'll
        # create those connection operations ourselves.
        omit = [*omit, 'code', 'data']

        # Gather the operations used to perform the basic ``stream_eq``...
        operations = super().stream_eq(interface, omit=omit, **kwargs)

        # ... and then add the operations necessary to connect our data/ctrl,
        # with endianness swaps.
        payload_words = len(self.ctrl)
        for i in range(payload_words):

            # Figure out what word we want to grab from, on the RHS.
            # It only matters that this is the word opposite of the word we're reading in the LHS.
            rhs_word_index = (payload_words - i) -1

            # Create the operations necessary to perform our assignment with our endian swap...
            endian_swap_operations = [
                self.data.word_select(i, 8)  .eq(interface.data.word_select(rhs_word_index, 8)),
                self.ctrl[i]                 .eq(interface.ctrl[rhs_word_index])
            ]

            #... and add it to our overall list of operations.
            operations.extend(endian_swap_operations)


        return operations



class SuperSpeedStreamArbiter(StreamArbiter):
    """ Convenience variant of our StreamArbiter that operates SuperSpeed streams in the ``ss`` domain. """

    def __init__(self):
        super().__init__(stream_type=USBRawSuperSpeedStream, domain="ss")


class SuperSpeedStreamInterface(StreamInterface):
    """ Convenience variant of our StreamInterface sized to work with SuperSpeed streams. """

    def __init__(self):
        super().__init__(payload_width=32, valid_width=4)



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Endpoint interfaces for working with streams.

The endpoint interfaces in this module provide endpoint interfaces suitable for
connecting streams to USB endpoints.
"""

from amaranth       import Elaboratable, Module, Signal

from ..endpoint     import EndpointInterface
from ...stream      import StreamInterface, USBOutStreamBoundaryDetector
from ..transfer     import USBInTransferManager
from ....memory     import TransactionalizedFIFO


class USBStreamInEndpoint(Elaboratable):
    """ Endpoint interface that transmits a simple data stream to a host.

    This interface is suitable for a single bulk or interrupt endpoint.

    This endpoint interface will automatically generate ZLPs when a stream packet would end without
    a short data packet. If the stream's ``last`` signal is tied to zero, then a continuous stream of
    maximum-length-packets will be sent with no inserted ZLPs.

    The ``flush`` input may be asserted to to cause all pending data to be transmitted as soon as
    possible. When ``flush`` is asserted, packets of varying length will be sent as needed, according
    to the data available.

    This implementation is double buffered; and can store a single packets worth of data while transmitting
    a second packet.


    Attributes
    ----------
    stream: StreamInterface, input stream
        Full-featured stream interface that carries the data we'll transmit to the host.

    flush: Signal(), input
        Assert to cause all pending data to be transmitted as soon as possible.

    discard: Signal(), input
        Assert to cause all pending data to be discarded.

    interface: EndpointInterface
        Communications link to our USB device.


    Parameters
    ----------
    endpoint_number: int
        The endpoint number (not address) this endpoint should respond to.
    max_packet_size: int
        The maximum packet size for this endpoint. Should match the wMaxPacketSize provided in the
        USB endpoint descriptor.
    """


    def __init__(self, *, endpoint_number, max_packet_size):

        self._endpoint_number = endpoint_number
        self._max_packet_size = max_packet_size

        #
        # I/O port
        #
        self.stream    = StreamInterface()
        self.interface = EndpointInterface()
        self.flush     = Signal()
        self.discard   = Signal()


    def elaborate(self, platform):
        m = Module()
        interface = self.interface

        # Create our transfer manager, which will be used to sequence packet transfers for our stream.
        m.submodules.tx_manager = tx_manager = USBInTransferManager(self._max_packet_size)

        # Check there has been a ClearFeature(ENDPOINT_HALT) request address to this endpoint.
        clear_endpoint_halt = \
            interface.clear_endpoint_halt_in.enable & \
            interface.clear_endpoint_halt_in.direction & \
            (interface.clear_endpoint_halt_in.number == self._endpoint_number)
        m.d.comb += [

            # Always generate ZLPs; in order to pass along when stream packets terminate.
            tx_manager.generate_zlps    .eq(1),

            # We want to handle packets only that target our endpoint number.
            tx_manager.active           .eq(interface.tokenizer.endpoint == self._endpoint_number),

            # Connect up our transfer manager to our input stream, flush and discard control...
            tx_manager.transfer_stream  .stream_eq(self.stream),
            tx_manager.flush            .eq(self.flush),
            tx_manager.discard          .eq(self.discard),

            # ... and data-toggle reset on clear endpoint halt...
            tx_manager.reset_sequence   .eq(clear_endpoint_halt),

            # ... and our output stream...
            interface.tx                .stream_eq(tx_manager.packet_stream),
            interface.tx_pid_toggle     .eq(tx_manager.data_pid),

            # ... and connect through our token/handshake signals.
            interface.tokenizer         .connect(tx_manager.tokenizer),
            tx_manager.handshakes_out   .connect(interface.handshakes_out),
            interface.handshakes_in     .connect(tx_manager.handshakes_in)
        ]

        return m



class USBMultibyteStreamInEndpoint(Elaboratable):
    """ Endpoint interface that transmits a simple data stream to a host.

    This interface is suitable for a single bulk or interrupt endpoint.

    This variant accepts streams with payload sizes that are a multiple of one byte; data is always
    transmitted to the host in little-endian byte order.

    This endpoint interface will automatically generate ZLPs when a stream packet would end without
    a short data packet. If the stream's ``last`` signal is tied to zero, then a continuous stream of
    maximum-length-packets will be sent with no inserted ZLPs.

    This implementation is double buffered; and can store a single packets worth of data while transmitting
    a second packet.


    Attributes
    ----------
    stream: StreamInterface, input stream
        Full-featured stream interface that carries the data we'll transmit to the host.
    interface: EndpointInterface
        Communications link to our USB device.


    Parameters
    ----------
    byte_width: int
        The number of bytes to be accepted at once.
    endpoint_number: int
        The endpoint number (not address) this endpoint should respond to.
    max_packet_size: int
        The maximum packet size for this endpoint. Should match the wMaxPacketSize provided in the
        USB endpoint descriptor.
    """
    def __init__(self, *, byte_width, endpoint_number, max_packet_size):
        self._byte_width      = byte_width
        self._endpoint_number = endpoint_number
        self._max_packet_size = max_packet_size

        #
        # I/O port
        #
        self.stream          = StreamInterface(payload_width=byte_width * 8)
        self.interface       = EndpointInterface()


    def elaborate(self, platform):
        m = Module()

        # Create our core, single-byte-wide endpoint, and attach it directly to our interface.
        m.submodules.stream_ep = stream_ep = USBStreamInEndpoint(
            endpoint_number=self._endpoint_number,
            max_packet_size=self._max_packet_size
        )
        stream_ep.interface = self.interface

        # Create semantic aliases for byte-wise and word-wise streams;
        # so the code below reads more clearly.
        byte_stream = stream_ep.stream
        word_stream = self.stream

        # We'll put each word to be sent through an shift register
        # that shifts out words a byte at a time.
        data_shift = Signal.like(word_stream.payload)

        # Latched versions of our first and last signals.
        first_latched = Signal()
        last_latched  = Signal()

        # Count how many bytes we have left to send.
        bytes_to_send = Signal(range(0, self._byte_width + 1))

        # Always provide our inner transmitter with the least byte of our shift register.
        m.d.comb += byte_stream.payload.eq(data_shift[0:8])


        with m.FSM(domain="usb"):

            # IDLE: transmitter is waiting for input
            with m.State("IDLE"):
                m.d.comb += word_stream.ready.eq(1)

                # Once we get a send request, fill in our shift register, and start shifting.
                with m.If(word_stream.valid):
                    m.d.usb += [
                        data_shift         .eq(word_stream.payload),
                        first_latched      .eq(word_stream.first),
                        last_latched       .eq(word_stream.last),

                        bytes_to_send      .eq(self._byte_width - 1),
                    ]
                    m.next = "TRANSMIT"


            # TRANSMIT: actively send each of the bytes of our word
            with m.State("TRANSMIT"):
                m.d.comb += byte_stream.valid.eq(1)

                # Once the byte-stream is accepting our input...
                with m.If(byte_stream.ready):
                    is_first_byte = (bytes_to_send == self._byte_width - 1)
                    is_last_byte  = (bytes_to_send == 0)

                    # Pass through our First and Last signals, but only on the first and
                    # last bytes of our word, respectively.
                    m.d.comb += [
                        byte_stream.first  .eq(first_latched & is_first_byte),
                        byte_stream.last   .eq(last_latched  & is_last_byte)
                    ]

                    # ... if we have bytes left to send, move to the next one.
                    with m.If(bytes_to_send > 0):
                        m.d.usb += [
                            bytes_to_send .eq(bytes_to_send - 1),
                            data_shift    .eq(data_shift[8:]),
                        ]

                    # Otherwise, complete the frame.
                    with m.Else():
                        m.d.comb += word_stream.ready.eq(1)

                        # If we still have data to send, move to the next byte...
                        with m.If(self.stream.valid):
                            m.d.usb += [
                                data_shift     .eq(word_stream.payload),
                                first_latched  .eq(word_stream.first),
                                last_latched   .eq(word_stream.last),

                                bytes_to_send  .eq(self._byte_width - 1),
                            ]

                        # ... otherwise, move to our idle state.
                        with m.Else():
                            m.next = "IDLE"


        return m



class USBStreamOutEndpoint(Elaboratable):
    """ Endpoint interface that receives data from the host, and produces a simple data stream.

    This interface is suitable for a single bulk or interrupt endpoint.


    Attributes
    ----------
    stream: StreamInterface, output stream
        Full-featured stream interface that carries the data we've received from the host.
    interface: EndpointInterface
        Communications link to our USB device.

    Parameters
    ----------
    endpoint_number: int
        The endpoint number (not address) this endpoint should respond to.
    max_packet_size: int
        The maximum packet size for this endpoint. If this there isn't `max_packet_size` space in
        the endpoint buffer, this endpoint will NAK (or participate in the PING protocol.)
    buffer_size: int, optional
        The total amount of data we'll keep in the buffer; typically two max-packet-sizes or more.
        Defaults to twice the maximum packet size.
    """


    def __init__(self, *, endpoint_number, max_packet_size, buffer_size=None):
        self._endpoint_number = endpoint_number
        self._max_packet_size = max_packet_size
        self._buffer_size = buffer_size if (buffer_size is not None) else (self._max_packet_size * 2 - 1)

        #
        # I/O port
        #
        self.stream    = StreamInterface()
        self.interface = EndpointInterface()


    def elaborate(self, platform):
        m = Module()

        stream    = self.stream
        interface = self.interface
        tokenizer = interface.tokenizer

        #
        # Internal state.
        #

        # Stores the data toggle value we expect.
        expected_data_toggle = Signal()

        # Stores whether we've had a receive overflow.
        overflow = Signal()

        # Stores a count of received bytes in the current packet.
        rx_cnt = Signal(range(self._max_packet_size))

        # Stores whether we're in the middle of a transfer.
        transfer_active = Signal()

        #
        # Receiver logic.
        #

        # Create a version of our receive stream that has added `first` and `last` signals, which we'll use
        # internally as our main stream.
        m.submodules.boundary_detector = boundary_detector = USBOutStreamBoundaryDetector()
        m.d.comb += [
            interface.rx                   .stream_eq(boundary_detector.unprocessed_stream),
            boundary_detector.complete_in  .eq(interface.rx_complete),
            boundary_detector.invalid_in   .eq(interface.rx_invalid),
        ]

        rx       = boundary_detector.processed_stream
        rx_first = boundary_detector.first
        rx_last  = boundary_detector.last

        # Create a Rx FIFO.
        m.submodules.fifo = fifo = TransactionalizedFIFO(width=10, depth=self._buffer_size, name="rx_fifo", domain="usb")


        #
        # Create some basic conditionals that will help us make decisions.
        #

        endpoint_number_matches  = (tokenizer.endpoint == self._endpoint_number)
        targeting_endpoint       = endpoint_number_matches & tokenizer.is_out

        expected_pid_match       = (interface.rx_pid_toggle == expected_data_toggle)
        sufficient_space         = (fifo.space_available >= self._max_packet_size)

        ping_response_requested  = endpoint_number_matches & tokenizer.is_ping & tokenizer.ready_for_response
        data_response_requested  = targeting_endpoint & tokenizer.is_out & interface.rx_ready_for_response

        okay_to_receive          = targeting_endpoint & expected_pid_match
        data_is_lost             = okay_to_receive & rx.next & rx.valid & fifo.full
        data_accepted            = okay_to_receive & ~data_is_lost & ~overflow
        should_skip              = targeting_endpoint & ~expected_pid_match

        full_packet              = rx_cnt == self._max_packet_size - 1

        m.d.comb += [

            # We'll always populate our FIFO directly from the receive stream; but we'll also include our
            # "short packet detected" signal, as this indicates that we're detecting the last byte of a transfer.
            fifo.write_data[0:8] .eq(rx.payload),
            fifo.write_data[8]   .eq(rx_last & ~full_packet),
            fifo.write_data[9]   .eq(rx_first & ~transfer_active),
            fifo.write_en        .eq(okay_to_receive & rx.next & rx.valid & ~fifo.full),

            # We'll keep data if our packet finishes with a valid CRC and no overflow; and discard it otherwise.
            fifo.write_commit    .eq(targeting_endpoint & boundary_detector.complete_out & ~overflow),
            fifo.write_discard   .eq(targeting_endpoint & (boundary_detector.invalid_out | (boundary_detector.complete_out & overflow))),

            # We'll ACK each packet if it's received correctly; _or_ if we skipped the packet
            # due to a PID sequence mismatch. If we get a PID sequence mismatch, we assume that
            # we missed a previous ACK from the host; and ACK without accepting data [USB 2.0: 8.6.3].
            interface.handshakes_out.ack  .eq(
                (data_response_requested & data_accepted) |
                (ping_response_requested & sufficient_space) |
                (data_response_requested & should_skip)
            ),

            # We'll NAK any time we want to accept a packet, but we don't have enough room.
            interface.handshakes_out.nak  .eq(
                (data_response_requested & ~data_accepted & ~should_skip) |
                (ping_response_requested & ~sufficient_space)
            ),

            # Our stream data always comes directly out of the FIFO; and is valid
            # whenever our FIFO actually has data for us to read.
            stream.valid      .eq(~fifo.empty),
            stream.payload    .eq(fifo.read_data[0:8]),

            # Our `last` bit comes directly from the FIFO; and we know a `first` bit immediately
            # follows a `last` one.
            stream.last       .eq(fifo.read_data[8]),
            stream.first      .eq(fifo.read_data[9]),

            # Move to the next byte in the FIFO whenever our stream is advaced.
            fifo.read_en      .eq(stream.ready),
            fifo.read_commit  .eq(1)
        ]

        # Count bytes in packet.
        with m.If(fifo.write_en):
            m.d.usb += rx_cnt.eq(rx_cnt + 1)

            # Set the transfer active flag depending on whether this is a full packet.
            with m.If(rx_last):
                m.d.usb += transfer_active.eq(full_packet)

        # We'll set the overflow flag if we're receiving data we don't have room for.
        with m.If(data_is_lost):
            m.d.usb += overflow.eq(1)

        # We'll clear the overflow flag and byte counter when the packet is done.
        with m.Elif(fifo.write_commit | fifo.write_discard):
            m.d.usb += overflow.eq(0)
            m.d.usb += rx_cnt.eq(0)

        # We'll toggle our DATA PID each time we issue an ACK to the host [USB 2.0: 8.6.2].
        with m.If(data_response_requested & data_accepted):
            m.d.usb += expected_data_toggle.eq(~expected_data_toggle)

        # If there has been a ClearFeature(ENDPOINT_HALT) request address to this endpoint...
        clear_endpoint_halt = \
            self.interface.clear_endpoint_halt_in.enable & \
            ~self.interface.clear_endpoint_halt_in.direction & \
            (self.interface.clear_endpoint_halt_in.number == self._endpoint_number)

        with m.If(clear_endpoint_halt):
            # ... reset the expected data toggle.
            m.d.usb += expected_data_toggle.eq(0)


        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test       import LunaUSBGatewareTestCase, usb_domain_test_case

from luna.gateware.usb.stream import USBOutStreamBoundaryDetector

class USBOutStreamBoundaryDetectorTest(LunaUSBGatewareTestCase):
    FRAGMENT_UNDER_TEST   = USBOutStreamBoundaryDetector

    @usb_domain_test_case
    def test_boundary_detection(self):
        dut                 = self.dut
        processed_stream    = self.dut.processed_stream
        unprocesesed_stream = self.dut.unprocessed_stream

        # Before we see any data, we should have all of our strobes de-asserted, and an invalid stream.
        self.assertEqual((yield processed_stream.valid), 0)
        self.assertEqual((yield processed_stream.next), 0)
        self.assertEqual((yield dut.first), 0)
        self.assertEqual((yield dut.last), 0)

        # If our stream goes valid...
        yield unprocesesed_stream.valid.eq(1)
        yield unprocesesed_stream.next.eq(1)
        yield unprocesesed_stream.payload.eq(0xAA)
        yield

        # ... we shouldn't see anything this first cycle...
        self.assertEqual((yield processed_stream.valid), 0)
        self.assertEqual((yield processed_stream.next), 0)
        self.assertEqual((yield dut.first), 0)
        self.assertEqual((yield dut.last), 0)

        # ... but after two cycles...
        yield unprocesesed_stream.payload.eq(0xBB)
        yield
        yield unprocesesed_stream.payload.eq(0xCC)
        yield

        # ... we should see a valid stream's first byte.
        self.assertEqual((yield processed_stream.valid), 1)
        self.assertEqual((yield processed_stream.next),  1)
        self.assertEqual((yield processed_stream.payload),  0xAA)
        self.assertEqual((yield dut.first), 1)
        self.assertEqual((yield dut.last), 0)
        yield unprocesesed_stream.payload.eq(0xDD)

        # ... followed by a byte that's neither first nor last...
        yield
        self.assertEqual((yield processed_stream.payload),  0xBB)
        self.assertEqual((yield dut.first), 0)
        self.assertEqual((yield dut.last), 0)

        # Once our stream is no longer valid...
        yield unprocesesed_stream.valid.eq(0)
        yield unprocesesed_stream.next.eq(0)
        yield
        yield

        # ... we should see our final byte.
        self.assertEqual((yield processed_stream.payload),  0xDD)
        self.assertEqual((yield dut.first), 0)
        self.assertEqual((yield dut.last), 1)




#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test                   import LunaSSGatewareTestCase, ss_domain_test_case

from luna.gateware.usb.usb3.physical.lfps import LFPSGenerator, _PollingLFPS, _PollingLFPSBurst, _PollingLFPSRepeat
from math import ceil

class LFPSGeneratorTest(LunaSSGatewareTestCase):
    FRAGMENT_UNDER_TEST = LFPSGenerator
    FRAGMENT_ARGUMENTS  = dict(
        lfps_pattern = _PollingLFPS,
        sys_clk_freq = 125e6
    )

    @ss_domain_test_case
    def test_polling_lfps_sequence(self):
        dut = self.dut

        burst_length = ceil(self.SS_CLOCK_FREQUENCY * _PollingLFPSBurst.t_typ)
        burst_repeat = ceil(self.SS_CLOCK_FREQUENCY * _PollingLFPSRepeat.t_typ)

        # Trigger a burst...
        yield dut.generate.eq(1)
        yield
        yield
        yield dut.generate.eq(0)

        # Wait for a whole burst-repeat cycle...
        burst_ticks = 0
        total_ticks = 0
        while (yield dut.drive_electrical_idle):

            # ... and measure how long our burst lasts...
            if (yield dut.send_signaling):
                burst_ticks += 1

            # ... as well as the length of our whole interval.
            total_ticks += 1
            yield

        # Our observed burst length should be within 10% of our specification...
        self.assertLess(abs(burst_ticks)/burst_length - 1.0, 10e-2)

        # ... as should our observed total length between bursts.
        self.assertLess(abs(total_ticks)/burst_repeat - 1.0, 10e-2)




#
# This file is part of LUNA.
#
# Copyright (c) 2020-2025 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Endpoint interfaces for isochronous endpoints.

These interfaces provide interfaces for connecting streams or stream-like
interfaces to hosts via isochronous pipes.
"""

from amaranth              import *
from amaranth.lib          import stream, wiring
from amaranth.lib .wiring  import In, Out

from ..endpoint            import EndpointInterface
from ...stream             import USBOutStreamBoundaryDetector
from ....stream.future     import Packet
from ....memory            import TransactionalizedFIFO


class USBIsochronousStreamOutEndpoint(Elaboratable):
    """ Endpoint interface that receives isochronous data from the host, and produces a simple data stream.

    Used for repeatedly streaming data from a host to a stream or stream-like interface.
    Intended to be useful as a transport for e.g. video or audio data.


    Attributes
    ----------
    stream: StreamInterface, output stream
        Full-featured stream interface that carries the data we've received from the host.
    interface: EndpointInterface
        Communications link to our USB device.

    Parameters
    ----------
    endpoint_number: int
        The endpoint number (not address) this endpoint should respond to.
    max_packet_size: int, optional
        The maximum packet size for this endpoint. If there isn't `max_packet_size` space in
        the endpoint buffer, additional data will be silently dropped.
    buffer_size: int, optional
        The total amount of data we'll keep in the buffer; typically two (TODO three?) max-packet-sizes or more.
        Defaults to twice (TODO three?) times the maximum packet size.
    """

    def __init__(self, *, endpoint_number, max_packet_size, buffer_size=None):
        self._endpoint_number = endpoint_number
        self._max_packet_size = max_packet_size
        # TODO self._buffer_size = buffer_size if (buffer_size is not None) else (self._max_packet_size * 3)
        self._buffer_size = buffer_size if (buffer_size is not None) else (self._max_packet_size * 2)

        #
        # I/O port
        #
        self.stream = stream.Interface(
            stream.Signature(
                Packet(unsigned(8))
            )
        )
        self.interface = EndpointInterface()

    def elaborate(self, platform):
        m = Module()

        stream    = self.stream
        interface = self.interface
        tokenizer = interface.tokenizer

        #
        # Internal state.
        #

        # Stores whether we've had a receive overflow.
        overflow = Signal()

        # Stores a count of received bytes in the current packet.
        rx_cnt = Signal(range(self._max_packet_size))

        #
        # Receiver logic.
        #

        # Create a version of our receive stream that has added `first` and `last` signals, which we'll use
        # internally as our main stream.
        m.submodules.boundary_detector = boundary_detector = USBOutStreamBoundaryDetector()
        m.d.comb += [
            interface.rx                   .stream_eq(boundary_detector.unprocessed_stream),
            boundary_detector.complete_in  .eq(interface.rx_complete),
            boundary_detector.invalid_in   .eq(interface.rx_invalid),
        ]

        rx       = boundary_detector.processed_stream
        rx_first = boundary_detector.first
        rx_last  = boundary_detector.last

        # Create a Rx FIFO.
        m.submodules.fifo = fifo = TransactionalizedFIFO(width=10, depth=self._buffer_size, name="rx_fifo", domain="usb")

        #
        # Create some basic conditionals that will help us make decisions.
        #

        endpoint_number_matches  = (tokenizer.endpoint == self._endpoint_number)
        targeting_endpoint       = endpoint_number_matches & tokenizer.is_out

        sufficient_space         = (fifo.space_available >= self._max_packet_size)

        okay_to_receive          = targeting_endpoint & sufficient_space
        data_is_lost             = okay_to_receive & rx.next & rx.valid & fifo.full

        full_packet              = rx_cnt == self._max_packet_size - 1

        m.d.comb += [

            # We'll always populate our FIFO directly from the receive stream; but we'll also include our
            # "short packet detected" signal, as this indicates that we're detecting the last byte of a transfer.
            fifo.write_data[0:8] .eq(rx.payload),
            fifo.write_data[8]   .eq(rx_last),
            fifo.write_data[9]   .eq(rx_first),
            fifo.write_en        .eq(okay_to_receive & rx.next & rx.valid),

            # We'll keep data if our packet finishes with a valid CRC; and discard it otherwise.
            fifo.write_commit    .eq(targeting_endpoint & boundary_detector.complete_out),
            fifo.write_discard   .eq(targeting_endpoint & boundary_detector.invalid_out),

            # Our stream data always comes directly out of the FIFO; and is valid
            # whenever our FIFO actually has data for us to read.
            stream.valid        .eq(~fifo.empty),
            stream.p.data       .eq(fifo.read_data[0:8]),

            # Our `last` bit comes directly from the FIFO; and we know a `first` bit immediately
            # follows a `last` one.
            stream.p.last        .eq(fifo.read_data[8]),
            stream.p.first       .eq(fifo.read_data[9]),

            # Move to the next byte in the FIFO whenever our stream is advanced.
            fifo.read_en         .eq(stream.ready),
            fifo.read_commit     .eq(1)
        ]

        # Count bytes in packet.
        with m.If(fifo.write_en):
            m.d.usb += rx_cnt.eq(rx_cnt + 1)

        # We'll set the overflow flag if we're receiving data we don't have room for.
        with m.If(data_is_lost):
            m.d.usb += overflow.eq(1)

        # We'll clear the overflow flag and byte counter when the packet is done.
        with m.Elif(fifo.write_commit | fifo.write_discard):
            m.d.usb += overflow.eq(0)
            m.d.usb += rx_cnt.eq(0)

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# Copyright (c) 2020 Florent Kermarrec <florent@enjoy-digital.fr>
#
# Code adapted from ``litex`` and ``usb3_pipe``.
# SPDX-License-Identifier: BSD-3-Clause
""" Code for USB3 physical-layer encoding. """

from amaranth import *

def K(x, y):
    """ Converts K/control codes to bytes

    Does not result in 8b10b data; requires explicit encoding.
    See the USB3 / PCIe specifications for more information.
    """
    return (y << 5) | x

def D(x, y):
    """ Converts D/data codes to bytes.

    Does not result in 8b10b data; requires explicit encoding.
    See the USB3 / PCIe specifications for more information.
    """
    return (y << 5) | x


class NamedSymbol:
    """ Simple encapsulation of a USB3 symbol, with simple metadata. """

    def __init__(self, name, value, description="", is_data=False):
        self.name        = name
        self.value       = value
        self.description = description
        self.ctrl        = 0 if is_data else 1

    def value_const(self, *, repeat=1):
        """ Returns this symbol's data value as an Amaranth const. """
        value = Const(self.value, 8)
        return value.replicate(repeat)


    def ctrl_const(self, *, repeat=1):
        """ Returns this symbol's ctrl value as an Amaranth const. """
        ctrl =  Const(self.ctrl, 1)
        return ctrl.replicate(repeat)


SKP =  NamedSymbol("SKP", K(28, 1), "Skip")                              # 3c
SDP =  NamedSymbol("SDP", K(28, 2), "Start Data Packet")                 # 5c
EDB =  NamedSymbol("EDB", K(28, 3), "End Bad")                           # 7c
SUB =  NamedSymbol("SUB", K(28, 4), "Decode Error Substitution")         # 9c
COM =  NamedSymbol("COM", K(28, 5), "Comma")                             # bc
RSD =  NamedSymbol("RSD", K(28, 6), "Reserved")                          # dc
SHP =  NamedSymbol("SHP", K(27, 7), "Start Header Packet")               # fb
END =  NamedSymbol("END", K(29, 7), "End")                               # fd
SLC =  NamedSymbol("SLC", K(30, 7), "Start Link Command")                # fe
EPF =  NamedSymbol("EPF", K(23, 7), "End Packet Framing")                # f7
IDL =  NamedSymbol("IDL", D(0, 0),  "Logical Idle", is_data=True)        # 00

symbols = [SKP, SDP, EDB, SUB, COM, RSD, SHP, END, SLC, EPF]


def get_word_for_symbols(*target_symbols):
    """ Returns a pair of Amaranth constants containing the data and ctrl values for the given symbols. """

    # Create constants that match the target data/ctrl bits for the given set of symbols.
    target_data = Cat(symbol.value_const() for symbol in target_symbols)
    target_ctrl = Cat(symbol.ctrl_const() for  symbol in target_symbols)

    return target_data, target_ctrl


def stream_matches_symbols(stream, *target_symbols, include_ready=False):
    """ Returns an Amaranth conditional that evaluates true when a stream contains the given four symbols.

    Notes:
        - The given conditional evaluates to False when ``stream.valid`` is falsey.
        - Assumes the stream is little endian, so the bytes of the stream would read SYM3 SYM2 SYM1 SYM0.
    """

    target_data, target_ctrl = get_word_for_symbols(*target_symbols)
    stream_ready = stream.ready if include_ready else True

    return (
        stream.valid & stream_ready  &
        (stream.data == target_data) &
        (stream.ctrl == target_ctrl)
    )


def stream_word_matches_symbol(stream, word_number, *, symbol):
    """ Returns an Amaranth conditional that evaluates true if the given word of a stream matches the given symbol. """

    return (
        stream.valid &
        (stream.data.word_select(word_number, 8) == symbol.value) &
        (stream.ctrl[word_number] == symbol.ctrl)
    )




# amaranth: UnusedElaboratable=no
#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test         import LunaGatewareTestCase, sync_test_case
from unittest       import TestCase

from amaranth import Module, Signal
from luna.gateware.utils.cdc import stretch_strobe_signal


class StrobeStretcherTest(LunaGatewareTestCase):
    """ Test case for our strobe stretcher function. """


    def instantiate_dut(self):
        m = Module()

        # Create a module that only has our stretched strobe signal.
        m.strobe_in = Signal()
        m.stretched_strobe = stretch_strobe_signal(m, m.strobe_in, to_cycles=2)

        return m


    def initialize_signals(self):
        yield self.dut.strobe_in.eq(0)


    @sync_test_case
    def test_stretch(self):

        # Ensure our stretched strobe stays 0 until it sees an input.
        yield
        self.assertEqual((yield self.dut.stretched_strobe), 0)
        yield
        self.assertEqual((yield self.dut.stretched_strobe), 0)

        # Apply our strobe, and validate that we immediately see a '1'...
        yield self.dut.strobe_in.eq(1)
        yield
        self.assertEqual((yield self.dut.stretched_strobe), 1)

        # ... ensure that 1 lasts for a second cycle ...
        yield self.dut.strobe_in.eq(0)
        yield
        self.assertEqual((yield self.dut.stretched_strobe), 1)

        # ... and then returns to 0.
        yield
        self.assertEqual((yield self.dut.stretched_strobe), 0)

        yield
        self.assertEqual((yield self.dut.stretched_strobe), 0)



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.interface.spi import SPIGatewareTestCase
from luna.gateware.test import LunaGatewareTestCase, sync_test_case

from amaranth import Signal
from luna.gateware.interface.i2c import I2CBus, I2CInitiator

class I2CInitiatorTestbench(I2CInitiator):
    def __init__(self, pads, period_cyc, clk_stretch=True):
        super().__init__(pads, period_cyc, clk_stretch)
        self.scl_o = Signal(init=1)  # used to override values from testbench
        self.sda_o = Signal(init=1)  

    def elaborate(self, platform):
        m = super().elaborate(platform)
        m.d.comb += [
            self.bus.scl_t.i.eq((self.bus.scl_t.o | ~self.bus.scl_t.oe) & self.scl_o),
            self.bus.sda_t.i.eq((self.bus.sda_t.o | ~self.bus.sda_t.oe) & self.sda_o),
        ]
        return m


class TestI2CInitiator(LunaGatewareTestCase):
    FRAGMENT_UNDER_TEST = I2CInitiatorTestbench
    FRAGMENT_ARGUMENTS = { "pads": I2CBus(), "period_cyc": 16 }

    def wait_condition(self, strobe):
        yield from self.wait_until(strobe, timeout=3*self.dut.period_cyc)

    def start(self):
        yield from self.pulse(self.dut.start)
        yield from self.wait_condition(self.dut.bus.start)

    def stop(self):
        yield from self.pulse(self.dut.stop)
        yield from self.wait_condition(self.dut.bus.stop)
    
    @sync_test_case
    def test_start(self):
        yield from self.start()
        self.assertEqual((yield self.dut.busy), 0)

    @sync_test_case
    def test_repeated_start(self):
        yield self.dut.bus.sda_o.eq(0)
        yield
        yield
        yield from self.start()
        yield from self.wait_condition(self.dut.bus.start)
        self.assertEqual((yield self.dut.busy), 0)

    @sync_test_case
    def test_stop(self):
        yield self.dut.bus.sda_o.eq(0)
        yield
        yield
        yield from self.stop()
        self.assertEqual((yield self.dut.busy), 0)

    def write(self, data, bits, ack):
        yield self.dut.data_i.eq(data)
        yield from self.pulse(self.dut.write)
        for n, bit in enumerate(bits):
            yield
            yield
            yield from self.wait_condition(self.dut.bus.scl_i == 0)
            yield from self.wait_condition(self.dut.bus.scl_i == 1)
            self.assertEqual((yield self.dut.bus.sda_i), bit)
            yield
        yield from self.advance_cycles(self.dut.period_cyc // 2)
        yield
        yield
        yield from self.wait_condition(self.dut.bus.scl_i == 0)
        yield self.dut.sda_o.eq(not ack)
        yield from self.wait_condition(self.dut.bus.scl_i == 1)
        yield self.dut.sda_o.eq(1)
        self.assertEqual((yield self.dut.busy), 1)
        yield from self.advance_cycles(self.dut.period_cyc // 2)
        yield
        yield
        yield
        yield
        self.assertEqual((yield self.dut.busy), 0)
        self.assertEqual((yield self.dut.ack_o), ack)

    @sync_test_case
    def test_write_ack(self):
        yield self.dut.bus.sda_o.eq(0)
        yield
        yield
        yield from self.write(0xA5, [1, 0, 1, 0, 0, 1, 0, 1], 1)

    @sync_test_case
    def test_write_nak(self):
        yield self.dut.bus.sda_o.eq(0)
        yield
        yield
        yield from self.write(0x5A, [0, 1, 0, 1, 1, 0, 1, 0], 0)

    @sync_test_case
    def test_write_tx(self):
        yield from self.start()
        yield from self.write(0x55, [0, 1, 0, 1, 0, 1, 0, 1], 1)
        yield from self.write(0x33, [0, 0, 1, 1, 0, 0, 1, 1], 0)
        yield from self.stop()
        yield
        yield
        self.assertEqual((yield self.dut.bus.sda_i), 1)
        self.assertEqual((yield self.dut.bus.scl_i), 1)

    def read(self, data, bits, ack):
        yield self.dut.ack_i.eq(ack)
        yield from self.pulse(self.dut.read)
        for n, bit in enumerate(bits):
            yield
            yield
            yield from self.wait_condition(self.dut.bus.scl_i == 0)
            yield self.dut.sda_o.eq(bit)
            yield from self.wait_condition(self.dut.bus.scl_i == 1)
            yield
        yield self.dut.sda_o.eq(1)
        yield from self.advance_cycles(self.dut.period_cyc // 2)
        yield
        yield
        yield from self.wait_condition(self.dut.bus.scl_i == 0)
        yield from self.wait_condition(self.dut.bus.scl_i == 1)
        self.assertEqual((yield self.dut.bus.sda_i), not ack)
        self.assertEqual((yield self.dut.busy), 1)
        yield from self.advance_cycles(self.dut.period_cyc // 2)
        yield
        yield
        yield
        yield
        self.assertEqual((yield self.dut.busy), 0)
        self.assertEqual((yield self.dut.data_o), data)

    @sync_test_case
    def test_read_ack(self):
        yield self.dut.bus.sda_o.eq(0)
        yield
        yield
        yield from self.read(0xA5, [1, 0, 1, 0, 0, 1, 0, 1], 1)

    @sync_test_case
    def test_read_nak(self):
        yield self.dut.bus.sda_o.eq(0)
        yield
        yield
        yield from self.read(0x5A, [0, 1, 0, 1, 1, 0, 1, 0], 0)

    @sync_test_case
    def test_read_tx(self):
        yield from self.start()
        yield from self.read(0x55, [0, 1, 0, 1, 0, 1, 0, 1], 1)
        yield from self.read(0x33, [0, 0, 1, 1, 0, 0, 1, 1], 0)
        yield from self.stop()
        yield
        yield
        self.assertEqual((yield self.dut.bus.sda_i), 1)
        self.assertEqual((yield self.dut.bus.scl_i), 1)



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test    import LunaSSGatewareTestCase, ss_domain_test_case

from luna.gateware.usb.usb3.link.data import DataPacketReceiver

class DataPacketReceiverTest(LunaSSGatewareTestCase):
    FRAGMENT_UNDER_TEST = DataPacketReceiver

    def initialize_signals(self):
        yield self.dut.sink.valid.eq(1)

    def provide_data(self, *tuples):
        """ Provides the receiver with a sequence of (data, ctrl) values. """

        # Provide each word of our data to our receiver...
        for data, ctrl in tuples:
            yield self.dut.sink.data.eq(data)
            yield self.dut.sink.ctrl.eq(ctrl)
            yield


    @ss_domain_test_case
    def test_unaligned_1B_packet_receive(self):

        # Provide a packet pair to the device.
        # (This data is from an actual recorded data packet.)
        yield from self.provide_data(
            # Header packet.
            # data       ctrl
            (0xF7FBFBFB, 0b1111),
            (0x32000008, 0b0000),
            (0x00010000, 0b0000),
            (0x08000000, 0b0000),
            (0xE801A822, 0b0000),

            # Payload packet.
            (0xF75C5C5C, 0b1111),
            (0x000000FF, 0b0000),
            (0xFDFDFDFF, 0b1110),
        )

        self.assertEqual((yield self.dut.packet_good), 1)


    @ss_domain_test_case
    def test_unaligned_2B_packet_receive(self):

        # Provide a packet pair to the device.
        # (This data is from an actual recorded data packet.)
        yield from self.provide_data(
            # Header packet.
            # data       ctrl
            (0xF7FBFBFB, 0b1111),
            (0x34000008, 0b0000),
            (0x00020000, 0b0000),
            (0x08000000, 0b0000),
            (0xD005A242, 0b0000),

            # Payload packet.
            (0xF75C5C5C, 0b1111),
            (0x2C98BBAA, 0b0000),
            (0xFDFD4982, 0b1110),
        )

        self.assertEqual((yield self.dut.packet_good), 1)



    @ss_domain_test_case
    def test_aligned_packet_receive(self):

        # Provide a packet pair to the device.
        # (This data is from an actual recorded data packet.)
        yield from self.provide_data(
            # Header packet.
            # data       ctrl
            (0xF7FBFBFB, 0b1111),
            (0x00000008, 0b0000),
            (0x00088000, 0b0000),
            (0x08000000, 0b0000),
            (0xA8023E0F, 0b0000),

            # Payload packet.
            (0xF75C5C5C, 0b1111),
            (0x001E0500, 0b0000),
            (0x00000000, 0b0000),
            (0x0EC69325, 0b0000),
        )

        self.assertEqual((yield self.dut.packet_good), 1)




#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test              import LunaSSGatewareTestCase, ss_domain_test_case

from luna.gateware.usb.usb3.link.crc import DataPacketPayloadCRC

class DataPacketPayloadCRCTest(LunaSSGatewareTestCase):
    FRAGMENT_UNDER_TEST = DataPacketPayloadCRC

    @ss_domain_test_case
    def test_aligned_crc(self):
        dut = self.dut

        #yield dut.advance_word.eq(1)

        for i in (0x02000112, 0x40000000):
            yield dut.data_input.eq(i)
            yield from self.pulse(dut.advance_word, step_after=False)

        self.assertEqual((yield dut.crc), 0x34984B13)


    @ss_domain_test_case
    def test_unaligned_crc(self):
        dut = self.dut


        # Aligned section of a real USB data capture, from a USB flash drive.
        aligned_section =[
            0x03000112,
            0x09000000,
            0x520013FE,
            0x02010100,
        ]

        # Present the aligned section...
        for i in aligned_section:
            yield dut.data_input.eq(i)
            yield from self.pulse(dut.advance_word, step_after=False)

        # ... and then our unaligned data.
        yield dut.data_input.eq(0x0000_0103)
        yield

        # Our next-CRC should indicate the correct value...
        self.assertEqual((yield dut.next_crc_2B), 0x540aa487)

        # ...and after advancing, we should see the same value on our CRC output.
        yield from self.pulse(dut.advance_2B)
        self.assertEqual((yield dut.crc), 0x540aa487)



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" USB3 physical-layer abstraction."""

import logging

from amaranth import *
from amaranth.lib.cdc import PulseSynchronizer, FFSynchronizer


class PHYResetController(Elaboratable):
    """ Gateware responsible for bringing up a PIPE PHY.

    Note that this gateware resides in `sync`, rather than one of our
    SuperSpeed domains, as the SuperSpeed PHY has yet to bring up its clocks.

    Attributes
    ----------
    reset: Signal(), output
        Signal that drives the PHY's ``reset`` signal.

    phy_status: Signal(), input
        The PIPE PHY's phy_status signal; tracks our progress in startup.

    ready: Signal(), output
        Status signal; asserted when the PHY has started up and is ready for use.

    Parameters
    ----------
    sync_frequency: float
        The frequency of the sync clock domain.

    """

    def __init__(self, *, sync_frequency):
        self._sync_frequency = sync_frequency

        #
        # I/O port
        #
        self.reset          = Signal()
        self.ready          = Signal()

        self.phy_status     = Signal()



    def elaborate(self, platform):
        m = Module()

        # Keep track of how many cycles we'll keep our PHY in reset.
        # This is larger than any requirement, in order to work with a broad swathe of PHYs,
        # in case a PHY other than the TUSB1310A ever makes it to the market.
        cycles_in_reset = int(5e-6 * 50e6)
        cycles_spent_in_reset = Signal(range(cycles_in_reset + 1))


        with m.FSM():

            # STARTUP_RESET -- post configuration, we'll reset the PIPE PHY.
            # This is distinct from the PHY's built-in power-on-reset, as we run this
            # on every FPGA configuration.
            with m.State("STARTUP_RESET"):
                m.d.comb += [
                    self.reset        .eq(1),
                ]

                # Once we've extended past a reset time, we can move on.
                m.d.sync += cycles_spent_in_reset.eq(cycles_spent_in_reset + 1)
                with m.If(cycles_spent_in_reset == cycles_in_reset):
                    m.next = "DETECT_PHY_STARTUP"


            # DETECT_PHY_STARTUP -- post-reset, the PHY should drive its status line high.
            # We'll wait for this to happen, so we can track the PHY's progress.
            with m.State("DETECT_PHY_STARTUP"):

                with m.If(self.phy_status):
                    m.next = "WAIT_FOR_STARTUP"


            # WAIT_FOR_STARTUP -- we've now detected that the PHY is starting up.
            # We'll wait for that startup signal to be de-asserted, indicating that the PHY is ready.
            with m.State("WAIT_FOR_STARTUP"):

                # For now, we'll start up in P0. This will change once we implement proper RxDetect.
                with m.If(~self.phy_status):
                    m.next = "READY"


            # READY -- our PHY is all started up and ready for use.
            # For now, we'll remain here until we're reset.
            with m.State("READY"):
                m.d.comb += self.ready.eq(1)


        return m



class LinkPartnerDetector(Elaboratable):
    """ Light abstraction over our PIPE receiver detection mechanism.

    Primarily responsible for the power state sequencing necessary during receiver detection.

    Attributes
    ----------
    request_detection: Signal(), input
        Strobe; requests that a receiver detection will be performed.

    power_state: Signal(2), output
        Controls the PHY's power state signals.
    detection_control: Signal(), output
        Controls the PHY's partner detection signal.
    phy_status: Signal(), input
        Status signal; asserted when the PHY has completed our request.
    rx_status: Signal(3), input
        Status signal; indicates the result of our request.

    new_result: Signal(), output
        Strobe; indicates when a new result is ready on :attr:``partner_present``.
    partner_present: Signal(), output
        High iff a partner was detected during the last detection cycle.

    Parameters
    ----------
    rx_status: Array(Signal(3), Signal(3))
        Read-only view of the PHY's rx_status signal.
    """

    def __init__(self):

        #
        # I/O port
        #
        self.request_detection = Signal()

        self.power_state       = Signal(2, init=2)
        self.detection_control = Signal()
        self.phy_status        = Signal()
        self.rx_status         = Signal(3)

        self.new_result        = Signal()
        self.partner_present   = Signal()



    def elaborate(self, platform):
        m = Module()

        # Partner detection is indicated by the value `011` being present on RX_STATUS
        # after a detection completes.
        PARTNER_PRESENT_STATUS = 0b011

        # Keep track of the previous value of the request detection strobe
        past_request_detection = Signal()
        m.d.ss                += past_request_detection.eq(self.request_detection)

        with m.FSM(domain="ss"):

            # IDLE_P2 -- our post-startup state; represents when we're IDLE but in P2.
            # This is typically only seen at board startup.
            with m.State("IDLE_P2"):
                m.d.comb += self.power_state.eq(2)

                with m.If(self.request_detection):
                    m.next = "PERFORM_DETECT"


            # PERFORM_DETECT -- we're asking our PHY to perform the core of our detection,
            # and waiting for that detection to complete.
            with m.State("PERFORM_DETECT"):

                # Per [TUSB1310A, 5.3.5.2], we should hold our detection control high until
                # PhyStatus pulses high; when we'll get the results of our detection.
                m.d.comb += [
                    self.power_state        .eq(2),
                    self.detection_control  .eq(1)
                ]

                # When we see PhyStatus strobe high, we know our result is in RxStatus.
                for i in range(2):

                    # When our detection is complete...
                    with m.If(self.phy_status):

                        # ... capture the results, but don't mark ourselves as complete, yet, as we're
                        # still in P2. We'll need to move to operational state.
                        m.d.ss += self.partner_present.eq(self.rx_status == PARTNER_PRESENT_STATUS)
                        m.next = "MOVE_TO_P0"


            # MOVE_TO_P0 -- we've completed a detection, and now are ready to move (back) into our
            # operational state.
            with m.State("MOVE_TO_P0"):

                # Ask the PHY to put us back down in P0.
                m.d.comb += self.power_state.eq(0)

                # Once the PHY indicates it's put us into the relevant power state, we're done.
                # We can now broadcast our result.
                with m.If(self.phy_status):
                    m.d.comb += self.new_result.eq(1)
                    m.next = "IDLE_P0"


            # IDLE_P0 -- our normal operational state; usually reached after at least one detection
            # has completed successfully. We'll wait until another detection is requested.
            with m.State("IDLE_P0"):
                m.d.comb += self.power_state.eq(0)

                # We can only perform detections from P2; so, when the user requests a detection, we'll
                # need to move back to P2.
                request_detection_rose = ~past_request_detection & self.request_detection
                with m.If(request_detection_rose):
                    m.next = "MOVE_TO_P2"


            # MOVE_TO_P2 -- our user has requested a detection, which we can only perform from P2.
            # Accordingly, we'll move to P2, and -then- perform our detection.
            with m.State("MOVE_TO_P2"):

                # Ask the PHY to put us into P2.
                m.d.comb += self.power_state.eq(2)

                # Once the PHY indicates it's put us into the relevant power state, we can begin
                # our link partner detection.
                with m.If(self.phy_status):
                    m.next = "PERFORM_DETECT"


        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Pre-made gateware that implements CDC-ACM serial. """

from amaranth                          import Elaboratable, Module, Signal

from ...stream                         import StreamInterface
from ..usb2.device                     import USBDevice
from ..usb2.request                    import USBRequestHandler, StallOnlyRequestHandler
from ..usb2.endpoints.stream           import USBStreamInEndpoint, USBStreamOutEndpoint

from usb_protocol.types                import USBRequestType
from usb_protocol.emitters             import DeviceDescriptorCollection
from usb_protocol.emitters.descriptors import cdc


class ACMRequestHandlers(USBRequestHandler):
    """ Minimal set of request handlers to implement ACM functionality.

    Implements just enough of the requests to be usable on major operating system.
    In testing, macOS and Linux are fine will all requests being stalled; while Windows
    seems to be happy as long as SET_LINE_CODING is implemented. We'll implement only
    that, and stall every other handler.
    """

    SET_LINE_CODING = 0x20

    def elaborate(self, platform):
        m = Module()

        interface         = self.interface
        setup             = self.interface.setup

        #
        # Class request handlers.
        #

        with m.If(setup.type == USBRequestType.CLASS):
            with m.Switch(setup.request):

                # SET_LINE_CODING: The host attempts to tell us how it wants serial data
                # encoding. Since we output a stream, we'll ignore the actual line coding.
                with m.Case(self.SET_LINE_CODING):

                    # Drive interface outputs for this request
                    m.d.comb += interface.claim.eq(1)

                    # Always ACK the data out...
                    with m.If(interface.rx_ready_for_response):
                        m.d.comb += interface.handshakes_out.ack.eq(1)

                    # ... and accept whatever the request was.
                    with m.If(interface.status_requested):
                        m.d.comb += self.send_zlp()

        return m


class USBSerialDevice(Elaboratable):
    """ Device that acts as a CDC-ACM 'serial converter'.

    Exposes a stream interface.

    Attributes
    ----------
    connect: Signal(), input
        When asserted, the USB-to-serial device will be presented to the host
        and allowed to communicate.
    rx: StreamInterface(), output stream
        A stream carrying data received from the host.
    tx: StreamInterface(), input stream
        A stream carrying data to be transmitted to the host.

    Parameters
    ----------
    bus: Record()
        The raw input record that provides our USB connection. Should be a connection to a USB PHY,
        SerDes, or raw USB lines as described at: https://luna.readthedocs.io/en/latest/custom_hardware.html.
    idVendor: int, <65536
        The Vendor ID that should be presented for the relevant USB device.
    idProduct: int, <65536
        The Product ID that should be presented for the relevant USB device.

    manufacturer_string: str, optional
        A string describing this device's manufacturer.
    product_str: str, optional
        A string describing this device.
    serial_number: str, optional
        A string describing this device's serial number.

    max_packet_size: int in {64, 246, 512}, optional
        The maximum packet size for communications.
    """

    _STATUS_ENDPOINT_NUMBER = 3
    _DATA_ENDPOINT_NUMBER   = 4

    def __init__(self, *, bus, idVendor, idProduct,
            manufacturer_string="LUNA",
            product_string="USB-to-serial",
            serial_number="", max_packet_size=64):

        self._bus                 = bus
        self._idVendor            = idVendor
        self._idProduct           = idProduct
        self._manufacturer_string = manufacturer_string
        self._product_string      = product_string
        self._serial_number       = serial_number
        self._max_packet_size     = max_packet_size

        #
        # I/O port
        #
        self.connect = Signal()
        self.rx      = StreamInterface()
        self.tx      = StreamInterface()


    def create_descriptors(self):
        """ Creates the descriptors that describe our serial topology. """

        descriptors = DeviceDescriptorCollection()

        # Create a device descriptor with our user parameters...
        with descriptors.DeviceDescriptor() as d:
            d.idVendor           = self._idVendor
            d.idProduct          = self._idProduct

            d.iManufacturer      = self._manufacturer_string
            d.iProduct           = self._product_string
            d.iSerialNumber      = self._serial_number

            d.bNumConfigurations = 1

        # ... and then describe our CDC-ACM setup.
        with descriptors.ConfigurationDescriptor() as c:
            # First, add a descriptor to show that both interfaces in this
            # configuration are associated with the same function.
            # (this seems to be required on Windows)
            with c.InterfaceAssociationDescriptor() as ia:
                ia.bFirstInterface = 0
                ia.bInterfaceCount = 2
                ia.bFunctionClass    = 0x02 # CDC
                ia.bFunctionSubClass = 0x02 # ACM
                ia.bFunctionProtocol = 0x01 # AT commands / UART

            # Then, we'll describe the Communication Interface, which contains most
            # of our description; but also an endpoint that does effectively nothing in
            # our case, since we don't have interrupts we want to send up to the host.
            with c.InterfaceDescriptor() as i:
                i.bInterfaceNumber   = 0

                i.bInterfaceClass    = 0x02 # CDC
                i.bInterfaceSubclass = 0x02 # ACM
                i.bInterfaceProtocol = 0x01 # AT commands / UART

                # Provide the default CDC version.
                i.add_subordinate_descriptor(cdc.HeaderDescriptorEmitter())

                # ... specify our interface associations ...
                union = cdc.UnionFunctionalDescriptorEmitter()
                union.bControlInterface      = 0
                union.bSubordinateInterface0 = 1
                i.add_subordinate_descriptor(union)

                # ... and specify the interface that'll carry our data...
                call_management = cdc.CallManagementFunctionalDescriptorEmitter()
                call_management.bDataInterface = 1
                i.add_subordinate_descriptor(call_management)

                # CDC communications endpoint
                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = 0x80 | self._STATUS_ENDPOINT_NUMBER
                    e.bmAttributes     = 0x03
                    e.wMaxPacketSize   = self._max_packet_size
                    e.bInterval        = 11

            # Finally, we'll describe the communications interface, which just has the
            # endpoints for our data in and out.
            with c.InterfaceDescriptor() as i:
                i.bInterfaceNumber   = 1
                i.bInterfaceClass    = 0x0a # CDC data
                i.bInterfaceSubclass = 0x00
                i.bInterfaceProtocol = 0x00

                # Data IN to host (tx, from our side)
                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = 0x80 | self._DATA_ENDPOINT_NUMBER
                    e.wMaxPacketSize   = self._max_packet_size

                # Data OUT from host (rx, from our side)
                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = self._DATA_ENDPOINT_NUMBER
                    e.wMaxPacketSize   = self._max_packet_size

        return descriptors


    def elaborate(self, platform):
        m = Module()

        # Create our core USB device, and add a standard control endpoint.
        m.submodules.usb = usb = USBDevice(bus=self._bus)
        control_ep = usb.add_standard_control_endpoint(self.create_descriptors())

        # Attach our class request handlers.
        control_ep.add_request_handler(ACMRequestHandlers())

        # Attach class-request handlers that stall any vendor or reserved requests,
        # as we don't have or need any.
        stall_condition = lambda setup : \
            (setup.type == USBRequestType.VENDOR) | \
            (setup.type == USBRequestType.RESERVED)
        control_ep.add_request_handler(StallOnlyRequestHandler(stall_condition))

        # Create our status/communications endpoint; but don't ever drive its stream.
        # This should be optimized down to an endpoint that always NAKs.
        serial_status_ep = USBStreamInEndpoint(
            endpoint_number=self._STATUS_ENDPOINT_NUMBER,
            max_packet_size=self._max_packet_size
        )
        usb.add_endpoint(serial_status_ep)

        # Create an endpoint for serial rx...
        serial_rx_endpoint = USBStreamOutEndpoint(
            endpoint_number=self._DATA_ENDPOINT_NUMBER,
            max_packet_size=self._max_packet_size,
        )
        usb.add_endpoint(serial_rx_endpoint)

        # ... and one for serial tx.
        serial_tx_endpoint = USBStreamInEndpoint(
            endpoint_number=self._DATA_ENDPOINT_NUMBER,
            max_packet_size=self._max_packet_size
        )
        usb.add_endpoint(serial_tx_endpoint)

        # Connect up our I/O.
        m.d.comb += [
            serial_tx_endpoint.stream  .stream_eq(self.tx),
            self.rx                    .stream_eq(serial_rx_endpoint.stream),
            usb.connect                .eq(self.connect)
        ]

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD--3-Clause
""" Endpoint interfaces for isochronous endpoints.

These interfaces provide interfaces for connecting memories or memory-like
interfaces to hosts via isochronous pipes.
"""

from amaranth       import Elaboratable, Module, Signal

from ..endpoint     import EndpointInterface


class USBIsochronousInEndpoint(Elaboratable):
    """ Isochronous endpoint that presents a memory-like interface.

    Used for repeatedly streaming data to a host from a memory or memory-like interface.
    Intended to be useful as a transport for e.g. video or audio data.

    Attributes
    ----------
    interface: EndpointInterface
        Communications link to our USB core.

    bytes_in_frame: Signal(range(0, 3073)), input
        Specifies how many bytes will be transferred during this frame. If this is 0,
        a single ZLP will be emitted; for any other value one, two, or three packets
        will be generated, depending on the packet size. Latched in at the start of
        each frame.

        The maximum allowed value for this signal depends on the number of transfers
        per (micro)frame:
        - If this is a high-speed, high-throughput endpoint (descriptor indicates
          maxPacketSize > 512 and multiple transfers per microframe), then this value
          maxes out at (N * maxPacketSize), where N is the number of transfers per microframe.
        - For all other configurations, this must be <= the maximum packet size.

    address: Signal(range(0,3072)), output
        Indicates the address / offset of the byte currently being transmitted.
        Can be used to drive the ``address` lines of an asynchronous memory
    next_address: Signal(range(0,3072)), output
        Indicates the "address" / offset of the byte that should be presented
        on :attr:``value`` at the next ``usb``-clock cycle. Can be used to drive
        the ``address`` lines of a synchronous memory.
    value: Signal(8), input
        The value to be transmitted, this cycle. Can be directly tied to the read
        port of a memory.

    Parameters
    ----------
    endpoint_number: int
        The endpoint number (not address) this endpoint should respond to.
    max_packet_size: int
        The maximum packet size for this endpoint. Should match the wMaxPacketSize provided in the
        USB endpoint descriptor.
    """

    _MAX_FRAME_DATA = 1024 * 3

    def __init__(self, *, endpoint_number, max_packet_size):
        self._endpoint_number = endpoint_number
        self._max_packet_size = max_packet_size

        #
        # I/O Port
        #
        self.interface      = EndpointInterface()

        self.bytes_in_frame = Signal(range(0, self._MAX_FRAME_DATA + 1))

        self.address        = Signal(range(0, self._MAX_FRAME_DATA))
        self.next_address   = Signal.like(self.address)
        self.value          = Signal(8)


    def elaborate(self, platform):
        m = Module()

        # Shortcuts.
        interface        = self.interface
        out_stream       = interface.tx
        new_frame        = interface.tokenizer.new_frame

        targeting_ep_num = (interface.tokenizer.endpoint == self._endpoint_number)
        targeting_us     = targeting_ep_num & interface.tokenizer.is_in
        data_requested   = targeting_us & interface.tokenizer.ready_for_response

        # Track our state in our transmission.
        bytes_left_in_frame  = Signal.like(self.bytes_in_frame)
        bytes_left_in_packet = Signal(range(0, self._max_packet_size + 1), init=self._max_packet_size - 1)
        next_data_pid        = Signal(2)

        # Reset our state at the start of each frame.
        with m.If(new_frame):

            m.d.usb += [
                # Latch in how many bytes we'll be transmitting this frame.
                bytes_left_in_frame.eq(self.bytes_in_frame),

                # And start with a full packet to transmit.
                bytes_left_in_packet.eq(self._max_packet_size)
            ]

            # If it'll take more than two packets to send our data, start off with DATA2.
            # We'll follow with DATA1 and DATA0.
            with m.If(self.bytes_in_frame > (2 * self._max_packet_size)):
                m.d.usb += next_data_pid.eq(2)

            # Otherwise, if we need two, start with DATA1.
            with m.Elif(self.bytes_in_frame > self._max_packet_size):
                m.d.usb += next_data_pid.eq(1)

            # Otherwise, we'll start (and end) with DATA0.
            with m.Else():
                m.d.usb += next_data_pid.eq(0)


        m.d.comb += [
            # Always pass our ``value`` directly through to our transmitter.
            # We'll provide ``address``/``next_address`` to our user code to help
            # orchestrate this timing.
            out_stream.payload       .eq(self.value),

            # Provide our data pid through to to the transmitter.
            interface.tx_pid_toggle  .eq(next_data_pid)
        ]

        #
        # Core sequencing FSM.
        #
        with m.FSM(domain="usb"):

            # IDLE -- the host hasn't yet requested data from our endpoint.
            with m.State("IDLE"):
                m.d.usb  += [
                    # Remain targeting the first byte in our frame.
                    self.address      .eq(0),
                    out_stream.first  .eq(0)
                ]

                m.d.comb += self.next_address.eq(0)

                # Once the host requests a packet from us...
                with m.If(data_requested):

                    # If we have data to send, send it.
                    with m.If(bytes_left_in_frame):
                        m.d.usb += out_stream.first.eq(1)
                        m.next = "SEND_DATA"

                    # Otherwise, we'll send a ZLP.
                    with m.Else():
                        m.next = "SEND_ZLP"


            # SEND_DATA -- our primary data-transmission state; handles packet transmission
            with m.State("SEND_DATA"):
                last_byte_in_packet    = (bytes_left_in_packet <= 1)
                last_byte_in_frame     = (bytes_left_in_frame  <= 1)
                byte_terminates_send   = last_byte_in_packet | last_byte_in_frame

                m.d.comb += [
                    # Our data is always valid in this state...
                    out_stream.valid .eq(1),

                    # ... and we're terminating our packet if we're on the last byte of it.
                    out_stream.last  .eq(byte_terminates_send),
                ]

                # ``address`` should always move to the value presented in
                # ``next_address`` on each clock edge.
                m.d.usb += self.address.eq(self.next_address)

                # By default, don't advance.
                m.d.comb += self.next_address.eq(self.address)

                # We'll advance each time our data is accepted.
                with m.If(out_stream.ready):
                    m.d.usb += out_stream.first.eq(0)

                    # Mark the relevant byte as sent...
                    m.d.usb += [
                        bytes_left_in_frame   .eq(bytes_left_in_frame  - 1),
                        bytes_left_in_packet  .eq(bytes_left_in_packet - 1),
                    ]

                    # ... and advance to the next address.
                    m.d.comb += self.next_address.eq(self.address + 1)

                    # If we've just completed transmitting a packet, or we've
                    # just transmitted a full frame, end our transmission.
                    with m.If(byte_terminates_send):
                        m.d.usb += [
                            # Move to the next DATA pid, which is always one DATA PID less.
                            # [USB2.0: 5.9.2]. We'll reset this back to its maximum value when
                            # the next frame starts.
                            next_data_pid        .eq(next_data_pid - 1),

                            # Mark our next packet as being a full one.
                            bytes_left_in_packet .eq(self._max_packet_size)
                        ]
                        m.next = "IDLE"


            # SEND_ZLP -- sends a zero-length packet, and then return to idle.
            with m.State("SEND_ZLP"):
                # We'll request a ZLP by strobing LAST and VALID without strobing FIRST.
                m.d.comb += [
                    out_stream.valid  .eq(1),
                    out_stream.last   .eq(1),
                ]
                m.next = "IDLE"

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Utilities for building USB descriptors into gateware. """

import struct
import functools

from amaranth                                import Cat, DomainRenamer, Elaboratable, Module, Signal
from amaranth.lib.memory                     import Memory
from usb_protocol.emitters.descriptors       import DeviceDescriptorCollection
from usb_protocol.types.descriptors.standard import StandardDescriptorNumbers

from ..stream                                import USBInStreamInterface
from ...stream.generator                     import ConstantStreamGenerator
from ...utils.bus                            import OneHotMultiplexer

class USBDescriptorStreamGenerator(ConstantStreamGenerator):
    """ Specialized stream generator for generating USB descriptor constants. """

    def __init__(self, data):
        """
        Parameters:
            descriptor_number -- The descriptor number represented.
            data              -- The raw bytes (or equivalent) for the descriptor.
        """

        # Always create USB descriptors in the USB domain; always have a maximum length field that can
        # be up to 16 bits wide, and always use USBInStream's. These allow us to tie easily to our request
        # handlers.
        super().__init__(data, domain="usb", stream_type=USBInStreamInterface, max_length_width=16)



class GetDescriptorHandlerDistributed(Elaboratable):
    """ Gateware that handles responding to GetDescriptor requests.

    Currently does not support descriptors in multiple languages.

    I/O port:
        I: value[16]  -- The value field associated with the Get Descriptor request.
                         Contains the descriptor type and index.
        I: length[16] -- The length field associated with the Get Descriptor request.
                         Determines the maximum amount allowed in a response.

        I: start      -- Strobe that indicates when a descriptor should be transmitted.

        *: tx         -- The USBInStreamInterface that streams our descriptor data.
        O: stall      -- Pulsed if a STALL handshake should be generated, instead of a response.
    """

    def __init__(self, descriptor_collection: DeviceDescriptorCollection, max_packet_length=64):
        """
        Parameteres:
            descriptor_collection -- The DeviceDescriptorCollection containing the descriptors
                                     to use for this device.
        """

        self._descriptors = descriptor_collection
        self._max_packet_length = max_packet_length

        #
        # I/O port
        #
        self.value          = Signal(16)
        self.length         = Signal(16)

        self.start          = Signal()
        self.start_position = Signal(11)

        self.tx             = USBInStreamInterface()
        self.stall          = Signal()


    def elaborate(self, platform):
        m = Module()

        # Collection that will store each of our descriptor-generation submodules.
        descriptor_generators = {}

        #
        # Figure out the maximum length we're willing to send.
        #
        length = Signal(16)

        # We'll never send more than our MaxPacketSize. This means that we'll want to send a maximum of
        # either our maximum packet length, or the amount of data we have remaining; whichever is less.
        #
        # Note that this doesn't take into account the length of the actual data to be sent; this is handled
        # in the stream generator.
        words_remaining = self.length - self.start_position
        with m.If(words_remaining <= self._max_packet_length):
            m.d.comb += length.eq(words_remaining)
        with m.Else():
            m.d.comb += length.eq(self._max_packet_length)


        #
        # Create our constant-stream generators for each of our descriptors.
        #
        for type_number, index, raw_descriptor in self._descriptors:
            # Create the generator...
            if isinstance(raw_descriptor, bytes):
                generator = USBDescriptorStreamGenerator(raw_descriptor)
            else:
                generator = raw_descriptor()
            descriptor_generators[(type_number, index)] = generator

            m.d.comb += [
                generator.max_length     .eq(length),
                generator.start_position .eq(self.start_position)
            ]

            # ... and attach it to this module.
            type_ref =  type_number.name if isinstance(type_number, StandardDescriptorNumbers) else type_number
            setattr(m.submodules, f'USBDescriptorStreamGenerator({type_ref},{index})', generator)


        #
        # Connect up each of our generators.
        #

        with m.Switch(self.value):

            # Generate a conditional interconnect for each of our items.
            for (type_number, index), generator in descriptor_generators.items():

                # If the value matches the given type number...
                with m.Case(type_number << 8 | index):

                    # ... connect the relevant generator to our output.
                    m.d.comb += generator.stream  .attach(self.tx)
                    m.d.usb += generator.start    .eq(self.start),

            # If none of our descriptors match, stall any request that comes in.
            with m.Default():
                m.d.comb += self.stall.eq(self.start)


        return m


class GetDescriptorHandlerBlock(Elaboratable):
    """ Gateware that handles responding to GetDescriptor requests.

    Currently does not support descriptors in multiple languages.

    I/O port:
        I: value[16]      -- The value field associated with the Get Descriptor request.
                             Contains the descriptor type and index.
        I: length[16]     -- The length field associated with the Get Descriptor request.
                             Determines the maximum amount allowed in a response.

        I: start          -- Strobe that indicates when a descriptor should be transmitted.
        I: start_position -- Specifies the starting position of the descriptor data to be transmitted.

        *: tx             -- The USBInStreamInterface that streams our descriptor data.
        O: stall          -- Pulsed if a STALL handshake should be generated, instead of a response.
    """

    ELEMENT_SIZE = 4

    COUNT_SIZE_BITS   = 16
    ADDRESS_SIZE_BITS = 16

    def __init__(self, descriptor_collection: DeviceDescriptorCollection, max_packet_length=64, domain="usb"):
        """
        Parameters
        ----------
        descriptor_collection: DeviceDescriptorCollection
            The DeviceDescriptorCollection containing the descriptors to use for this device.
        max_packet_length: int
            Maximum packet length.
        domain: string
            The clock domain this generator should belong to. Defaults to 'usb'.
        """

        self._descriptors        = descriptor_collection
        self._max_packet_length  = max_packet_length
        self._domain             = domain

        #
        # I/O port
        #
        self.value          = Signal(16)
        self.length         = Signal(16)

        self.start          = Signal()
        self.start_position = Signal(11)

        self.tx             = USBInStreamInterface()
        self.stall          = Signal()


    @classmethod
    def _align_to_element_size(cls, n):
        """ Returns a given number rounded up to the next "aligned" element size. """
        return (n + (cls.ELEMENT_SIZE - 1)) // cls.ELEMENT_SIZE

    def generate_rom_content(self):
        """ Generates the contents of the ROM used to hold descriptors.

        Memory layout
        -------------

        All data is aligned on 4 bytes

        Type offsets and number of entries
        ----------------------------------

        Each descriptor type starting from 0 until maximum used type number has
        an entry consisting of number of indexes for this type number (2 bytes)
        and address of first index (2 bytes).

        Invalid entries have a value of 0x0000xxxx (0 entries).

        Example:
        0000  0xFFFF
        0002  0xFFFF
        0004  Number of device indexes
        0006  Address of first device index
        0008  Number of configuration indexes
        000A  Address of first configuration index
        000C  Number of string indexes
        000E  Address of first string index
        ...

        Index offsets and length of descriptor
        --------------------------------------
        Each index of a descriptor type has an entry consistent of the length
        of the descriptor (2 bytes) and the address of first data byte (2 bytes).

        0010  Length of first device descriptor
        0012  Address of first device descriptor
        ...

        Data
        ----
        Descriptor data for each descriptor. Padded by 0 to next 4-byte address.

        ...   Descriptor data

        """

        # Get all descriptors and cache them in a dictionary, so that we can access them at will.
        descriptors = {}
        for type_number, index, raw_descriptor in self._descriptors:
            if type_number not in descriptors:
                descriptors[type_number] = {}

            descriptors[type_number][index] = raw_descriptor

        # Check if we need to support non-consecutive indexes.
        for type_number, indexes in sorted(descriptors.items()):
            if max(indexes.keys()) != len(indexes) - 1:
                indirect_idx = True
                break
        else:
            indirect_idx = False

        #
        # Compute the ROM size that we'll need.
        #
        max_type_number         = max(descriptors.keys())
        max_descriptor_size     = 0

        # Our ROM starts with a collection of pointers to our various descriptor tables...
        rom_size_table_pointers = (max_type_number + 1) * self.ELEMENT_SIZE

        # ... tables of pointers to each actual descriptor...
        table_entry_count = functools.reduce(lambda x, indexes: x + len(indexes), descriptors.values(), 0)
        rom_size_table_entries = table_entry_count * self.ELEMENT_SIZE

        # ... and the descriptors themselves.
        rom_size_descriptors = 0
        for descriptor_set in descriptors.values():
            for raw_descriptor in descriptor_set.values():

                # Compute the maximum size for each descriptor...
                aligned_size = self._align_to_element_size(len(raw_descriptor))
                rom_size_descriptors += aligned_size * self.ELEMENT_SIZE

                # ... and store the maximum size we've encountered.
                max_descriptor_size = max(max_descriptor_size, len(raw_descriptor))

        # Create an array to hold our initial values for our composite ROM.
        total_size = \
            rom_size_table_pointers + \
            rom_size_table_entries  + \
            rom_size_descriptors
        rom = bytearray(total_size)

        #
        # Fill the ROM's initialization values.
        #
        next_free_address       = (max_type_number + 1) * self.ELEMENT_SIZE
        type_index_base_address = [0] * (max_type_number + 1)

        # First, generate a list of "table pointers", which point to the address of each type, in memory.
        for type_number, indexes in sorted(descriptors.items()):

            # Create our table pointer entry, which should always point to the next free address...
            pointer_bytes = struct.pack(">HH", len(indexes), next_free_address)

            # ...add the pointer to our ROM...
            type_base_address = type_number * self.ELEMENT_SIZE
            rom[type_base_address:type_base_address + self.ELEMENT_SIZE] = pointer_bytes

            # ... store the base address, for our subsequent fill...
            type_index_base_address[type_number] = next_free_address

            #... and move to the next entry.
            next_free_address += len(indexes) * self.ELEMENT_SIZE


        index_map = {}

        # Next, create the tables themselves, which are filled with data pointers,
        # and add our descriptors to our memory.
        for type_number, descriptor_set in sorted(descriptors.items()):
            for i, (index, raw_descriptor) in enumerate(sorted(descriptor_set.items())):

                # Create our descriptor pointer entries...
                pointer_bytes = struct.pack(">HH", len(raw_descriptor), next_free_address)

                # ... figure out where in the ROM we're going to store the pointer ...
                index_base_address = type_index_base_address[type_number] + i * self.ELEMENT_SIZE

                # ... add the pointer...
                rom[index_base_address:index_base_address + 4] = pointer_bytes

                # ... and then store the descriptor itself to the pointer address.
                rom[next_free_address:next_free_address+len(raw_descriptor)] = raw_descriptor

                # Figure out the next free position for a descriptor.
                aligned_size = self._align_to_element_size(len(raw_descriptor))
                next_free_address += aligned_size * self.ELEMENT_SIZE

                # Store in the index map if needed.
                if indirect_idx:
                    index_map[index | (type_number << 8)] = i

        assert total_size == len(rom)


        #
        # Finally, convert our ROM into an initialization vector.
        #
        total_elements = total_size // self.ELEMENT_SIZE
        element_size = self.ELEMENT_SIZE

        # Chunk our ROM into a collection of entries...
        rom_entries = (rom[(element_size * i):(element_size * i) + element_size] for i in range(total_elements))

        # ... and then convert that into an initializer value in the format Amaranth ROMs like (integers).
        initializer = [struct.unpack(">I", rom_entry)[0] for rom_entry in rom_entries]

        return initializer, max_descriptor_size, max_type_number, index_map


    def elaborate(self, platform) -> Module:
        m = Module()

        # Aliases for type/index
        type_number = Signal(8)
        index = Signal(8)

        m.d.comb += [
            index.eq(self.value.word_select(0, 8)),
            type_number.eq(self.value.word_select(1, 8))
        ]

        #
        # Create the ROM that stores our descriptors...
        #
        rom_content, descriptor_max_length, max_type_index, index_map = self.generate_rom_content()

        m.submodules.rom = rom = Memory(shape=32, depth=len(rom_content), init=rom_content)
        rom_read_port = rom.read_port()

        # Create convenience aliases to the upper and lower half of the ROM.
        rom_upper_half = rom_read_port.data.word_select(1, 16)
        rom_lower_half = rom_read_port.data.word_select(0, 16)

        # All of our ROM's metadata is composed of elements formatted as (count, pointer).
        # Grab a quick reference to the ROM's upper half, which stores the count...
        rom_element_count    = rom_upper_half

        # ... and to the ROM's lower half, not counting the last two bits (which are always 0,
        # as our pointers are always aligned). This creates an element pointer counted in words,
        # instead of in bytes; and thus one compatible with our read_port addr.
        rom_element_pointer  = rom_read_port.data.bit_select(2, len(rom_read_port.addr))

        #
        # Figure out the maximum length we're willing to send.
        #
        length = Signal(16)

        # We'll never send more than our MaxPacketSize. This means that we'll want to send a maximum of
        # either our maximum packet length, or the amount of data we have remaining; whichever is less.
        #
        words_remaining = self.length - self.start_position
        with m.If(words_remaining <= self._max_packet_length):
            m.d.sync += length.eq(words_remaining)
        with m.Else():
            m.d.sync += length.eq(self._max_packet_length)

        # Register that stores our current position in the stream.
        # We still want to be able to store a position beyond bounds (+1),
        # this is required for descriptors length multiple of the maximum packet size.
        # Like this we do not overflow our position and are able to send a ZLP on the next request.
        position_in_stream = Signal(range(descriptor_max_length + 1))
        bytes_sent = Signal.like(length)

        # Registers that store descriptor length and data base address.
        descriptor_length = Signal(16)
        descriptor_data_base_address = Signal.like(rom_read_port.addr)

        # Track when we're on the first and last packet.
        on_first_packet = position_in_stream == self.start_position
        on_last_packet = \
            (position_in_stream == (descriptor_length - 1)) | \
            (bytes_sent + 1 >= length)

        #
        # Core transmit logic.
        #

        with m.FSM():

            # IDLE -- we're currently waiting to send a descriptor.
            with m.State('IDLE'):

                # Reset our data-sent count...
                m.d.sync += bytes_sent.eq(0)

                # ... and always prepare to read whatever descriptor type is requested.
                m.d.comb += rom_read_port.addr.eq(type_number)

                # Once we have a request to start transmitting...
                with m.If(self.start):
                    m.next = 'START'

            # START -- retiming state to allow construction of the length signal
            with m.State('START'):
                # ... and always prepare to read whatever descriptor type is requested.
                m.d.comb += rom_read_port.addr.eq(type_number)

                # ... apply our start position...
                m.d.sync += position_in_stream.eq(self.start_position),

                is_valid_type = (type_number <= max_type_index)

                # If we have a descriptor we're able to send, prepare to send it.
                with m.If(is_valid_type):
                    m.next = 'LOOKUP_TYPE'

                # Otherwise, stall the request immediately.
                with m.Else():
                    m.d.comb += self.stall.eq(1)
                    m.next = 'IDLE'

                # Handle index mapping for non-consecutive descriptors.
                if len(index_map) != 0:
                    descr_idx = Signal.like(index)
                    
                    with m.Switch(Cat(index, type_number)):
                        for orig_idx, remapped_idx in index_map.items():
                            with m.Case(orig_idx):
                                m.d.sync += descr_idx.eq(remapped_idx)
                        with m.Default():
                            m.d.sync += descr_idx.eq(0xFF)  # invalid index
                else:
                    descr_idx = index

            # LOOKUP_TYPE -- we're now ready to start sending a descriptor, but we've not yet fetched the
            # descriptor from memory. First, we'll need to find the location of the table that contains each
            # descriptor pointer.
            with m.State('LOOKUP_TYPE'):

                # Our previous state already selected the ROM word associated with our "table of tables";
                # meaning the ROM's read port now contains (count, table-pointer) for the relevant ROM type.

                # If the requested type is greater than the maximum type number the ROM encodes,
                # stall the request and return to idle.
                with m.If(descr_idx >= rom_element_count):
                    m.d.comb += self.stall.eq(1)
                    m.next = "IDLE"

                # Otherwise, look up the type data in the ROM; and then move on to finding the descriptor itself.
                with m.Else():
                    m.d.comb += rom_read_port.addr.eq(rom_element_pointer + descr_idx)
                    with m.If(length == 0):
                        m.next = 'SEND_ZLP'
                    with m.Else():
                        m.next = 'LOOKUP_DESCRIPTOR'


            # LOOKUP_DESCRIPTOR -- we've now fetched from ROM the location of the descriptor in memory.
            # We'll decode it, and then prepare to start sending the descriptor.
            # descriptor from memory. First, we'll need to find the location of the table that contains each
            # descriptor pointer.
            with m.State('LOOKUP_DESCRIPTOR'):

                # Point our descriptor at the first word in our descriptor, offset by our current position
                # in the stream...
                m.d.comb += rom_read_port.addr.eq((rom_read_port.data + position_in_stream) >> 2)

                # ... and register the position and shape of our descriptor in memory.
                m.d.sync += [
                    descriptor_data_base_address  .eq(rom_element_pointer),
                    descriptor_length             .eq(rom_element_count),
                ]

                # Our current position may point out of bounds in case our descriptor length is a multiple
                # of the maximum packet size. We must send a ZLP now so the host knows the previous
                # packet was the end of the descriptor.
                with m.If(position_in_stream >= rom_element_count):
                    m.next = 'SEND_ZLP'
                with m.Else():
                    m.next = 'SEND_DESCRIPTOR'


            # SEND_DESCRIPTOR -- we finally are actively streaming our descriptor; which we'll complete until
            # our descriptor is fully sent.
            with m.State('SEND_DESCRIPTOR'):
                word_in_stream = position_in_stream >> 2
                byte_in_stream = position_in_stream.bit_select(0, 2)

                m.d.comb += [
                    self.tx.valid       .eq(1),

                    # Always drive the stream from our current memory output...
                    rom_read_port.addr  .eq(descriptor_data_base_address + word_in_stream),
                    self.tx.payload     .eq(rom_read_port.data.word_select(~byte_in_stream, 8)),

                    # ... and base First and Last based on our current position in the stream.
                    self.tx.first       .eq(on_first_packet),
                    self.tx.last        .eq(on_last_packet)
                ]

                # Once a given word is accepted, we're ready to move on.
                with m.If(self.tx.ready):

                    # If we're not yet done, move to the next byte in the stream.
                    with m.If(~on_last_packet):
                        m.d.sync += [
                            position_in_stream  .eq(position_in_stream + 1),
                            bytes_sent          .eq(bytes_sent + 1),
                        ]
                        m.d.comb += rom_read_port.addr.eq(descriptor_data_base_address+(position_in_stream + 1).bit_select(2, len(position_in_stream) - 2)),

                    # Otherwise, we've finished! Return to IDLE.
                    with m.Else():
                        # Reset some values, might not be really required
                        m.d.sync += [
                            descriptor_length             .eq(0),
                            descriptor_data_base_address  .eq(0)
                        ]
                        m.next = 'IDLE'

            # SEND_ZLP -- we've had an empty descriptor request, or a request that ended on a packet boundary.
            # Send a zero-length packet to end the transaction.
            with m.State('SEND_ZLP'):
                m.d.comb += [
                    # Pulse `last` without `first` to indicate a ZLP.
                    self.tx.valid.eq(1),
                    self.tx.last .eq(1),
                ]
                m.next = 'IDLE'


        # Convert our sync domain to the domain requested by the user, if necessary.
        if self._domain != "sync":
            m = DomainRenamer({"sync": self._domain})(m)

        return m


class GetDescriptorHandlerMux(Elaboratable):
    def __init__(self, domain="usb"):
        self._domain        = domain
        self._handlers      = []
        #
        # I/O port
        #
        self.value          = Signal(16)
        self.length         = Signal(16)

        self.start          = Signal()
        self.start_position = Signal(11)

        self.tx             = USBInStreamInterface()
        self.stall          = Signal()

    def add_descriptor_handler(self, handler):
        self._handlers.append(handler)

    def elaborate(self, platform):
        m = Module()

        m.submodules += self._handlers

        # The multiplexer only stalls when all the underlying handlers do.
        stalled = [ Signal(name=f"stalled_{i}") for i in range(len(self._handlers)) ]
        m.d.comb += self.stall.eq(Cat(stalled).all())

        # Connect the inputs to the descriptor handlers and register any
        # stall for the current transaction.
        for i, handler in enumerate(self._handlers):
            m.d.comb += [
                handler.value           .eq(self.value),
                handler.length          .eq(self.length),
                handler.start           .eq(self.start),
                handler.start_position  .eq(self.start_position),
            ]
            stall_latch = Signal(name=f"stall_latch_{i}")
            m.d.comb += stalled[i].eq(handler.stall | stall_latch)
            with m.If(self.start | self.stall):
                m.d.sync += stall_latch.eq(0)
            with m.If(handler.stall & ~self.stall):
                m.d.sync += stall_latch.eq(1)

        # Connect up our transmit interface.
        m.submodules.tx_mux = tx_mux = OneHotMultiplexer(
            interface_type=USBInStreamInterface,
            mux_signals=('payload',),
            or_signals=('valid', 'first', 'last'),
            pass_signals=('ready',)
        )
        tx_mux.add_interfaces(handler.tx for handler in self._handlers)
        m.d.comb += self.tx.stream_eq(tx_mux.output)

        # Convert our sync domain to the domain requested by the user, if necessary.
        if self._domain != "sync":
            m = DomainRenamer({"sync": self._domain})(m)

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2021 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Hardware for communicating over various FPGAs' debug interfaces. """

from amaranth         import *
from amaranth.lib.cdc import FFSynchronizer, PulseSynchronizer
from amaranth.hdl.ast import ValueCastable
from amaranth.hdl.rec import DIR_FANIN, DIR_FANOUT

from ..utils        import falling_edge_detected, rising_edge_detected
from .spi           import SPIRegisterInterface

class ECP5DebugSPIBridge(Elaboratable, ValueCastable):
    """ Hardware that creates a virtual 'debug SPI' port, exposed over JTAG.

    WARNING: all signals are asynchronous to any internal clock domain -- they're
    synchronized to the JTAG scan chain clock. They should be synchronized before
    internal use.

    Attributes
    ----------
    sck: Signal(), output
        The serial clock used for JTAG transfers.
    sdi: Signal(), output
        The serial data received from TDI.
    sdo: Signal(), input
        The serial data to be transmitted. Used with the ER1 instruction.
    cs: Signal(), output
        Active high chip-select. Used with the ER1 instruction.

    sdo_alt: Signal(), input
        Alternate serial data to be transmitted. Used with the ER2 instruction.
    cs_alt: Signal(), output
        Alternate, active-high chip-select. Used with the ER2 instruction.
    """

    fields = ['sck', 'sdi', 'sdo', 'cs']
    layout = [
        ('sck', 1, DIR_FANIN),
        ('sdi', 1, DIR_FANIN),
        ('sdo', 1, DIR_FANOUT),
        ('cs',  1, DIR_FANIN),
    ]

    def __init__(self):

        #
        # I/O port
        #
        self.sck     = Signal()
        self.sdi     = Signal()
        self.sdo     = Signal()
        self.cs      = Signal()

        self.sdo_alt = Signal()
        self.cs_alt  = Signal()


    def elaborate(self, platform):
        m = Module()

        jtck   = Signal()
        jce1   = Signal()
        jce2   = Signal()
        jshift = Signal()


        # Instantiate our core JTAG interface, and hook it up to our signals.
        # This essentially grabs a connection to the ECP5's JTAG data chain when the ER1 or ER2
        # instructions are loaded into its instruction register.
        m.submodules.jtag = jtag = Instance("JTAGG",
            o_JTCK   = jtck,
            o_JTDI   = self.sdi,
            i_JTDO1  = self.sdo,
            i_JTDO2  = self.sdo_alt,
            o_JCE1   = jce1,
            o_JCE2   = jce2,
            o_JSHIFT = jshift
        )

        # As part of our switch to SPI, we'll want to create a clock that behaves according to a
        # sane SPI mode. Accordingly, we'll invert the JTAG clock to match the most common SPI CPOL.
        m.d.comb += self.sck.eq(~jtck)

        #
        # We'll need to keep our chip selects from going high prematurely, e.g. while the JTAG
        # clock is still running, but before we're presenting SPI data.
        #
        # Accordingly, we'll make our logic synchronous to the JTAG clock.
        # This is technically a hack, but it saves a bunch of logic.
        #

        # Create a clock domain clocked from our JTAG clock.
        m.domains.jtag = ClockDomain(local=True, clk_edge="neg")
        m.d.comb += ClockSignal("jtag").eq(jtck)

        # Create our chip selects. Note that these are based on two conditions:
        # - the JCE flags only go high when the right instruction is loaded; and
        # - the JSHIFT signal is only asserted when we're actually shifting data.
        m.d.jtag += [
            self.cs      .eq(jce1 & jshift),
            self.cs_alt  .eq(jce2 & jshift),
        ]

        return m


    #
    # Helpers that let us treat this object like a record, so it can be used
    # interchangeably with requested I/O objects.
    #

    @ValueCastable.lowermethod
    def as_value(self):
        return Record(self.layout)

    @ValueCastable.lowermethod
    def __getitem__(self, key):
        return {
            'sck': self.sck,
            'sdo': self.sdo,
            'sdi': self.sdi,
            'cs':  self.cs
        }[key]

    def shape(self):
        return self.as_value().shape()

    def _synchronize_(self, m, output, o_domain="sync", stages=2):
        """ Creates a synchronized copy of this interface's I/O. """

        # Synchronize our inputs...
        m.submodules += [
            FFSynchronizer(self.sck, output.sck, o_domain=o_domain, stages=stages),
            FFSynchronizer(self.sdi, output.sdi, o_domain=o_domain, stages=stages),
            FFSynchronizer(self.cs,  output.cs,  o_domain=o_domain, stages=stages),
        ]

        # ... and connect our output directly through.
        m.d.comb += self.sdo.eq(output.sdo)



class JTAGCommandInterface(Elaboratable):
    """ Interface that allow us to receive simple register-style commands over ECP5 JTAGG.

    This module works in an emulation of JTAG, except both instruction and data are shifted
    in the SHIFT-DR state. To shift an instruction, place the Lattice ER1 instruction into the
    JTAG IR, and then shift the instruction in as data. To shift data, place the Lattice ER2
    instruction into the JTAG IR, and then shift data normally.


    Attributes
    ----------
    command: Signal(command_size), output
        The command read from the SPI bus.
    command_ready: Signal(), output
        Strobes high to indicate a new command is ready.

    word_received: Signal(word_size), output
        The most recent word received.
    word_complete: Signal(), output
        Strobe indicating a new word is present on word_in.
    word_to_send: Signal(word_size), input
        The word to be transmitted; latched in on next word_complete and while cs is low
    """

    def __init__(self, command_size=8, word_size=32, output_domain="sync"):
        self.command_size   = command_size
        self.word_size      = word_size
        self._output_domain = output_domain

        #
        # I/O port.
        #

        # Command I/O.
        self.command        = Signal(self.command_size)
        self.command_ready  = Signal()

        # Data I/O
        self.word_received  = Signal(self.word_size)
        self.word_to_send   = Signal.like(self.word_received)
        self.word_complete  = Signal()

        # Status
        self.idle    = Signal()
        self.stalled = Signal()


    def elaborate(self, platform):
        m = Module()

        #
        # Instruction and data registers.
        #
        ir_size              = self.command_size + 1
        dr_size              = self.word_size + 1
        instruction_register = Signal(ir_size, init=(2 ** ir_size - 1))
        data_register        = Signal(dr_size, init=(2 ** dr_size - 1))

        #
        # JTAG interface.
        #

        jtag_clk = Signal()
        jtag_tdi = Signal()

        jtag_tdo_instruction = Signal()
        jtag_tdo_data        = Signal()
        jtag_ce_instruction  = Signal()
        jtag_ce_data         = Signal()
        jtag_in_shift_dr     = Signal()
        jtag_not_in_reset    = Signal()
        jtag_in_reset        = Signal()
        jtag_in_update_dr    = Signal()
        jtag_rti_instruction = Signal()

        # Instantiate our core JTAG interface, and hook it up to our signals.
        # This essentially grabs a connection to the ECP5's JTAG data chain when the ER1 or ER2
        # instructions are loaded into its instruction register.
        m.submodules.jtag =  Instance("JTAGG",
            o_JTCK    = jtag_clk,
            o_JTDI    = jtag_tdi,
            i_JTDO1   = jtag_tdo_instruction,
            i_JTDO2   = jtag_tdo_data,
            o_JCE1    = jtag_ce_instruction,
            o_JCE2    = jtag_ce_data,
            o_JSHIFT  = jtag_in_shift_dr,
            o_JRSTN   = jtag_not_in_reset,
            o_JUPDATE = jtag_in_update_dr,
            o_JRTI1   = jtag_rti_instruction,
        )
        m.d.comb += jtag_in_reset.eq(~jtag_not_in_reset)

        # Edges on the JTAGG signals line up directly with the JTCK rising edge,
        # so create a delayed version of the clock to sample them reliably.
        jtag_clk_delayed = Signal()
        m.d.sync += jtag_clk_delayed.eq(jtag_clk)

        # Detect a rising edge in jtag_clk.
        jtag_strobe = rising_edge_detected(m, jtag_clk_delayed)

        # Always output the end of our scan chains.
        m.d.comb += [
            jtag_tdo_instruction  .eq(instruction_register[0]),
            jtag_tdo_data         .eq(data_register[0]),
        ]

        with m.If(jtag_strobe):
            # Once we're actively shifting an instruction over JTAG, capture it.
            shifting_instruction = Signal()
            m.d.sync += shifting_instruction.eq(jtag_ce_instruction & jtag_in_shift_dr)
            with m.If(jtag_in_reset):
                m.d.sync += instruction_register.eq(instruction_register.reset)
            with m.Elif(shifting_instruction):
                m.d.sync += instruction_register.eq(Cat(instruction_register[1:], jtag_tdi))


            # Once we're actively shifting data over JTAG, capture it.
            shifting_data = Signal()
            m.d.sync += shifting_data.eq(jtag_ce_data & jtag_in_shift_dr)
            with m.If(jtag_in_reset):
                m.d.sync += data_register.eq(data_register.reset)
            with m.Elif(shifting_data):
                m.d.sync += data_register.eq(Cat(data_register[1:], jtag_tdi))
            with m.Elif(jtag_rti_instruction):
                m.d.sync += data_register.eq(self.word_to_send)


        # Create our event strobes.
        command_ready = Signal()
        data_ready    = Signal()

        # Connect up our "data/command ready" signals.
        m.d.comb += [
           command_ready  .eq(falling_edge_detected(m, shifting_instruction)),
           data_ready     .eq(falling_edge_detected(m, shifting_data)),
        ]

        # Latch our output data when new data is ready.
        with m.If(command_ready):
            m.d.sync += self.command.eq(instruction_register[1:])
        with m.If(data_ready):
            m.d.sync += self.word_received.eq(data_register[1:])

        # Create sync-domain versions of our data/command ready signals, which are delayed
        # one cycle from our internal ones, to coincide with the point at which our data is latched.
        m.d.sync += [
            self.command_ready  .eq(command_ready),
            self.word_complete  .eq(data_ready)
        ]

        return m



class JTAGRegisterInterface(SPIRegisterInterface):
    """ JTAG-carried version of our SPI register interface. """


    def __init__(self, address_size=15, register_size=32, default_read_value=0, support_size_autonegotiation=True):
        """
        Parameters:
            address_size       -- the size of an address, in bits; recommended to be one bit
                                  less than a binary number, as the write command is formed by adding a one-bit
                                  write flag to the start of every address
            register_size      -- The size of any given register, in bits.
            default_read_value -- The read value read from a non-existent or write-only register.

            support_size_autonegotiation --
                If set, register 0 is used as a size auto-negotiation register. Functionally equivalent to
                calling .support_size_autonegotiation(); see its documentation for details on autonegotiation.
        """

        self.address_size  = address_size
        self.register_size = register_size
        self.default_read_value  = default_read_value

        #
        # I/O port
        #
        self.idle    = Signal()
        self.stalled = Signal()

        #
        # Internal details.
        #

        # Instantiate an SPI command transciever submodule.
        self.interface = JTAGCommandInterface(command_size=address_size + 1, word_size=register_size)

        # Create a new, empty dictionary mapping registers to their signals.
        self.registers = {}

        # Create signals for each of our register control signals.
        self._is_write = Signal()
        self._address  = Signal(self.address_size)

        if support_size_autonegotiation:
            self.support_size_autonegotiation()


    def _connect_interface(self, m):
        pass



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# Copyright (c) 2020 Florent Kermarrec <florent@enjoy-digital.fr>
#
# Code based in part on ``usb3_pipe``.
# SPDX-License-Identifier: BSD-3-Clause
""" Low-frequency periodic signaling gateware.

LFPS is the first signaling to happen during the initialization of the USB3.0 link.

LFPS allows partners to exchange Out Of Band (OOB) controls/commands and consists of bursts where
a "slow" clock is generated (between 10M-50MHz) for a specific duration and with a specific repeat
period. After the burst, the transceiver is put in electrical idle mode (same electrical level on
P/N pairs while in nominal mode P/N pairs always have an opposite level):

Transceiver level/mode: _=0, -=1 x=electrical idle
|-_-_-_-xxxxxxxxxxxxxxxxxxxx|-_-_-_-xxxxxxxxxxxxxxxxxxxx|...
|<burst>                    |<burst>                    |...
|<-----repeat period------->|<-----repeat period------->|...

A LFPS pattern is identified by a burst duration and repeat period.

To be able generate and receive LFPS, a transceiver needs to be able put its TX in electrical idle
and to detect RX electrical idle.
"""

from math import ceil

from amaranth         import *
from amaranth.lib.cdc import FFSynchronizer

from ....utils       import rising_edge_detected


__all__ = ['LFPSTransceiver']


#
# LPFS timing "constants", and collection classes that represent them.
#

class LFPSTiming:
    """LPFS timings with typical, minimum and maximum timing values."""

    def __init__(self, t_typ=None, t_min=None, t_max=None):
        self.t_typ = t_typ
        self.t_min = t_min
        self.t_max = t_max
        assert t_min is not None
        assert t_max is not None

        self.range = (t_min, t_max)


class LFPS:
    """LPFS patterns with burst and repeat timings."""

    def __init__(self, burst, repeat=None, cycles=None):
        self.burst  = burst
        self.repeat = repeat
        self.cycles = None


# Our actual pattern constants; as specified by the USB3 specification.
# [USB 3.2r1: Table 6-30]

_PollingLFPSBurst  = LFPSTiming(t_typ=1.0e-6,  t_min=0.6e-6, t_max=1.4e-6)
_PollingLFPSRepeat = LFPSTiming(t_typ=10.0e-6, t_min=6.0e-6, t_max=14.0e-6)
_PollingLFPS       = LFPS(burst=_PollingLFPSBurst, repeat=_PollingLFPSRepeat)

_ResetLFPSBurst    = LFPSTiming(t_typ=100.0e-3, t_min=80.0e-3,  t_max=120.0e-3)
_ResetLFPS         = LFPS(burst=_ResetLFPSBurst)


#
# Gateware for generating and detecting bursts of LFPS patterns. Does not deal with
# the actual 10-50 MHz LFPS clock, delegating that to the PHY.
#

class LFPSDetector(Elaboratable):
    """ LFPS Signaling Detector

    Compares received (and demodulated) LFPS signaling with a specified pattern.

    Attributes
    ----------

    signaling_received: Signal(), input
        Held high when our PHY is detecting LFPS square waves.
    detect: Signal(), output
        Strobes high when a valid LFPS burst is detected.
    """
    def __init__(self, lfps_pattern, ss_clk_frequency=125e6):
        self._pattern              = lfps_pattern
        self._clock_frequency      = ss_clk_frequency

        #
        # I/O port
        #
        self.signaling_received = Signal() # i
        self.detect             = Signal() # o


    def elaborate(self, platform):
        m = Module()

        # Create an in-domain version of our square-wave-detector signal.
        present = Signal()
        m.submodules.present_cdc = FFSynchronizer(self.signaling_received, present, o_domain="ss")

        # Figure out how large of a counter we're going to need...
        burst_cycles_min    = ceil(self._clock_frequency * self._pattern.burst.t_min)
        burst_cycles_max    = ceil(self._clock_frequency * self._pattern.burst.t_max)

        # If we have a repeat interval, include it in our calculations.
        if self._pattern.repeat is not None:
            repeat_cycles_max   = ceil(self._clock_frequency * self._pattern.repeat.t_max)
            repeat_cycles_min   = ceil(self._clock_frequency * self._pattern.repeat.t_min)
            counter_max         = max(burst_cycles_max, repeat_cycles_max)
        else:
            counter_max         = burst_cycles_max

        # ... and create our counter.
        count = Signal(range(0, counter_max + 1))
        m.d.ss += count.eq(count + 1)

        # Keep track of whether our previous iteration matched; as we're typically in detecting
        # sequences of two correct LFPS cycles in a row.
        last_iteration_matched = Signal()

        #
        # Detector state machine.
        #
        with m.FSM(domain="ss"):

            # WAIT_FOR_NEXT_BURST -- we're not currently in a measurement; but are waiting for a
            # burst to begin, so we can perform a full measurement.
            with m.State("WAIT_FOR_NEXT_BURST"):
                m.d.ss += last_iteration_matched.eq(0)

                # If we've just seen the start of a burst, start measuring it.
                with m.If(rising_edge_detected(m, present, domain="ss")):
                    m.d.ss += count.eq(1),
                    m.next = "MEASURE_BURST"

            # MEASURE_BURST -- we're seeing something we believe to be a burst; and measuring its length.
            with m.State("MEASURE_BURST"):

                # Failing case: if our counter has gone longer than our maximum burst time, this isn't
                # a relevant burst. We'll wait for the next one.
                with m.If(count == burst_cycles_max):
                    m.next = 'WAIT_FOR_NEXT_BURST'

                # Once our burst is over, we'll need to decide if the burst matches our pattern.
                with m.If(~present):

                    # Failing case: if our burst is over, but we've not yet reached our minimum burst time,
                    # then this isn't a relevant burst. We'll wait for the next one.
                    with m.If(count < burst_cycles_min):
                        m.next = 'WAIT_FOR_NEXT_BURST'

                    # If our burst ended within a reasonable span, we can move on.
                    with m.Else():

                        # If we don't have a repeat interval, we're done!
                        if self._pattern.repeat is None:
                            m.d.comb += self.detect.eq(1)
                            m.next = "WAIT_FOR_NEXT_BURST"

                        # Otherwise, we'll need to check the repeat interval, as well.
                        else:
                            m.next = "MEASURE_REPEAT"

            if self._pattern.repeat is not None:

                # MEASURE_REPEAT -- we've just finished seeing a burst; and now we're measuring the gap between
                # successive bursts, which the USB specification calls the "repeat interval". [USB3.2r1: Fig 6-32]
                with m.State("MEASURE_REPEAT"):

                    # Failing case: if our counter has gone longer than our maximum burst time, this isn't
                    # a relevant burst. We'll wait for the next one.
                    with m.If(count == repeat_cycles_max):
                        m.next = 'WAIT_FOR_NEXT_BURST'

                    # Once we see another potential burst, we'll start our detection back from the top.
                    with m.If(present):
                        m.d.ss += count.eq(1)
                        m.next = 'MEASURE_BURST'

                        # If this lasted for a reasonable repeat interval, we've seen a correct burst!
                        with m.If(count >= repeat_cycles_min):

                            # Mark this as a correct iteration, and if the previous iteration was also
                            # a correct one, indicate that we've detected our output.
                            m.d.ss   += last_iteration_matched.eq(1)
                            m.d.comb += self.detect.eq(last_iteration_matched)

                        with m.Else():
                            m.d.ss   += last_iteration_matched.eq(0)

        return m



class LFPSGenerator(Elaboratable):
    """ LFPS Signaling Generator

    Transmits (to be modulated) LFPS signaling that follows a specified pattern.

    Attributes
    ----------

    generate: Signal(), input
        When asserted, continuously generates LFPS patterns.
    done: Signal(), output
        Strobes high every time a cycle is completed.

    drive_electrical_idle: Signal(), output
        Held high while cycles are being generated; during both burst and repeat intervals.
    send_signaling: Signal(), output
        Held high during a burst.
    """
    def __init__(self, lfps_pattern, sys_clk_freq):
        self._pattern         = lfps_pattern
        self._clock_frequency = sys_clk_freq

        #
        # I/O ports
        #
        self.generate               = Signal() # i
        self.completed              = Signal() # o
        self.drive_electrical_idle  = Signal() # o
        self.send_signaling         = Signal() # o


    def elaborate(self, platform):
        m = Module()

        # Compute the amount of cycles it takes to transmit the burst and reach
        # the end of the pattern...
        burst_cycles  = ceil(self._clock_frequency * self._pattern.burst.t_typ)
        repeat_cycles = ceil(self._clock_frequency * self._pattern.repeat.t_typ)

        # ... and create our cycle counter.
        count = Signal(range(0, repeat_cycles))
        m.d.ss += count.eq(count + 1)

        with m.FSM(domain="ss"):

            # IDLE -- wait for an LFPS burst request.
            with m.State("IDLE"):
                m.d.ss += count.eq(0)

                # Once we get one, start a burst.
                with m.If(self.generate):
                    m.d.comb += self.drive_electrical_idle.eq(1)
                    m.next = "BURST"

            # BURST -- transmit an LFPS burst for the duration of the burst interval.
            with m.State("BURST"):
                m.d.comb += self.drive_electrical_idle.eq(1)
                m.d.comb += self.send_signaling.eq(1)

                with m.If(count + 1 == burst_cycles):
                    m.next = "WAIT"

            # WAIT -- do nothing for the remaining part of the repeat interval.
            with m.State("WAIT"):
                m.d.comb += self.drive_electrical_idle.eq(1)

                with m.If(count + 1 == repeat_cycles):
                    m.d.comb += self.completed.eq(1)
                    m.next = "IDLE"

        return m


class LFPSTransceiver(Elaboratable):
    """ Low-Frequency Periodic Signaling (LFPS) Transciever

    Transmits and receives the LPFS sequences required for a USB 3.0 link.

    Attributes
    ----------

    drive_electrical_idle: Signal(), output
        Held high when our PHY should be either in Electrical Idle, or transmit LFPS waveforms.
    send_signaling: Signal(), output
        Held high when our PHY should be transmitting LFPS square waves.
    signaling_received: Signal(), input
        Should be asserted when our PHY is receiving LFPS square waves.

    send_polling: Signal(), input
        Strobe. When asserted, begins Polling LFPS.
    cycles_sent: Signal(16), output
        Incremented every time an LFPS cycle is completed.

    polling_detected: Signal(), output
        Strobes high when Polling LFPS is detected.
    reset_detected: Signal(), output
        Strobes high when Reset LFPS is detected.
    """

    def __init__(self, ss_clk_freq=125e6):
        self._clock_frequency      = ss_clk_freq

        #
        # I/O ports
        #
        self.drive_electrical_idle = Signal() # o
        self.send_signaling        = Signal() # o
        self.signaling_received    = Signal() # i

        # LFPS burst generation
        self.send_polling          = Signal() # i
        self.cycles_sent           = Signal(16) # o

        # LFPS burst reception
        self.polling_detected      = Signal() # o
        self.reset_detected        = Signal() # o


    def elaborate(self, platform):
        m = Module()

        #
        # LFPS Receivers.
        #
        m.submodules.polling_detector = polling_detector = LFPSDetector(_PollingLFPS, self._clock_frequency)
        m.d.comb += [
            polling_detector.signaling_received .eq(self.signaling_received),
            self.polling_detected               .eq(polling_detector.detect)
        ]

        m.submodules.reset_detector = reset_detector = LFPSDetector(_ResetLFPS, self._clock_frequency)
        m.d.comb += [
            reset_detector.signaling_received   .eq(self.signaling_received),
            self.reset_detected                 .eq(reset_detector.detect)
        ]

        #
        # LFPS Transmitter(s).
        #
        m.submodules.polling_generator = polling_generator = LFPSGenerator(_PollingLFPS, self._clock_frequency)
        m.d.comb += [
            polling_generator.generate  .eq(self.send_polling),
            self.drive_electrical_idle  .eq(polling_generator.drive_electrical_idle),
            self.send_signaling         .eq(polling_generator.send_signaling),
        ]

        with m.If(polling_generator.generate):
            with m.If(polling_generator.completed):
                m.d.ss += self.cycles_sent.eq(self.cycles_sent + 1)
        with m.Else():
            m.d.ss += self.cycles_sent.eq(0)

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# Copyright (c) 2020 Florent Kermarrec <florent@enjoy-digital.fr>
#
# Code based on ``litex`` and ``usb3_pipe``.
# SPDX-License-Identifier: BSD-3-Clause
""" Link training support gateware.
"""

from amaranth import *

from ..physical.coding import D, COM
from ...stream         import USBRawSuperSpeedStream

TS1_SET_ID = D(10, 2)
TS2_SET_ID = D(5,  2)

#
# Training set contents.
#
TSEQ_SET_DATA = [
    0xC017FFBC, # 3  - 0
    0x02E7B214, # 7  - 4
    0x286E7282, # 11 - 8
    0xBF6DBEA6, # 15 - 12
    0x4A4A4A4A, # 19 - 16
    0x4A4A4A4A, # 23 - 20
    0x4A4A4A4A, # 27 - 24
    0x4A4A4A4A, # 31 - 28
]

TS1_SET_DATA = [
    0xBCBCBCBC, # 3  - 0
    0x4A4A0000, # 7  - 4
    0x4A4A4A4A, # 19 - 16
    0x4A4A4A4A, # 23 - 20
]

INVERTED_TS1_SET_DATA = [
    0xBCBCBCBC, # 3  - 0
    0xB5B50000, # 7  - 4
    0xB5B5B5B5, # 19 - 16
    0xB5B5B5B5, # 23 - 20
]

TS2_SET_DATA = [
    0xBCBCBCBC, # 3  - 0
    0x45450000, # 7  - 4
    0x45454545, # 19 - 16
    0x45454545, # 23 - 20
]


class TSBurstDetector(Elaboratable):
    """ Simple Training Set detector; capable of detecting basic training sets.

    Parameters
    ----------
    sink: USBRawSuperSpeedStream(), input stream
        The sink to monitor for ordered sets; should be fed pre-descrambler data.
    detected: Signal(), output
        Strobe; pulses high when a burst of ordered sets has been received.
    """

    def __init__(self, *, set_data, first_word_ctrl=0b1111, sets_in_burst=1, include_config=False):
        self._set_data            = set_data
        self._first_word_ctrl     = first_word_ctrl
        self._detection_threshold = sets_in_burst
        self._include_config      = include_config

        #
        # I/O port
        #
        self.sink                 = USBRawSuperSpeedStream()
        self.detected             = Signal()

        if self._include_config:
            self.hot_reset            = Signal()
            self.loopback_requested   = Signal()
            self.scrambling_disabled  = Signal()


    def elaborate(self, platform):
        m = Module()
        m.d.ss += self.detected.eq(0)

        # Always accept data from our SerDes.
        m.d.comb += self.sink.ready.eq(1)

        # Counter that tracks how many consecutive sets we've received.
        consecutive_set_count = Signal(range(0, self._detection_threshold + 1))

        # Aliases.
        data  = self.sink.data
        ctrl  = self.sink.ctrl


        def advance_on_match(count, target_ctrl=0b0000, fail_state="NONE_DETECTED"):
            data_matches = (data == self._set_data[count])
            ctrl_matches = (ctrl == target_ctrl)

            # Once we have a valid word in our stream...
            with m.If(self.sink.valid):

                # ... advance if that word matches; or move to our "fail state" otherwise.
                with m.If(data_matches & ctrl_matches):
                    m.next = f"{count + 1}_DETECTED"
                with m.Else():
                    m.next = fail_state


        last_state_number = len(self._set_data)
        with m.FSM(domain="ss"):

            # NONE_DETECTED -- we haven't seen any parts of our ordered set;
            # we're waiting for the first one.
            with m.State("NONE_DETECTED"):
                m.d.ss += consecutive_set_count.eq(0)
                m.next = "WAIT_FOR_FIRST"

            # WAIT_FOR_FIRST -- we're waiting to see the first word of our sequence
            with m.State("WAIT_FOR_FIRST"):
                advance_on_match(0, target_ctrl=self._first_word_ctrl, fail_state="WAIT_FOR_FIRST")

            # 1_DETECTED -- we're parsing the first data word; which we'll do slightly differently,
            # as it can contain a variable configuration field.
            with m.State("1_DETECTED"):
                # If this set includes a configuration field, then we'll want to compare
                # our data with that set removed. Otherwise, we compare normally.
                data_masked  = (data & 0xffff0000) if self._include_config else data
                data_matches = (data_masked  == self._set_data[1])
                ctrl_matches = (ctrl         == 0)

                # Once we have a valid word in our stream...
                with m.If(self.sink.valid):

                    # ... advance if that word matches; or move to our "fail state" otherwise.
                    with m.If(data_matches & ctrl_matches):
                        m.next = f"2_DETECTED"

                        # If we're including a configuration field, parse it before we continue.
                        if self._include_config:
                            m.d.ss += [
                                # Bit 0 of Symbol 5 => Hot Reset
                                self.hot_reset           .eq(data.word_select(1, 8)[0]),

                                # Bit 2 of Symbol 5 => Requests Loopback Mode
                                self.loopback_requested  .eq(data.word_select(1, 8)[2]),

                                # Bit 3 of Symbol 5 = > Requests we not use scrambling.
                                self.scrambling_disabled .eq(data.word_select(1, 8)[3]),
                            ]

                    with m.Else():
                        m.next = "NONE_DETECTED"


            for i in range(2, last_state_number):
                with m.State(f"{i}_DETECTED"):
                    advance_on_match(i)


            with m.State(f"{last_state_number}_DETECTED"):

                # If we've seen as many sets as we're looking to detect, reset our count,
                # and indicate that we've completed a detection.
                with m.If(consecutive_set_count + 1 == self._detection_threshold):
                    m.d.ss   += [
                        consecutive_set_count  .eq(0),
                        self.detected          .eq(1)
                    ]

                # Otherwise, increase the number of consecutive sets seen.
                with m.Else():
                    m.d.ss += consecutive_set_count.eq(consecutive_set_count + 1)

                with m.If(self.sink.valid):
                    advance_on_match(0, target_ctrl=self._first_word_ctrl)
                with m.Else():
                    m.next = "WAIT_FOR_FIRST"


        return m



class TSEmitter(Elaboratable):
    """ Training Set Emitter

    Generic Training Sequence Ordered Set generator.

    This module generates a specific Training Sequence Ordered Set to the TX stream. For each start,
    N consecutive Ordered Sets are generated (N configured by n_ordered_sets). Done signal is assert
    on the last cycle of the generation. Training config can also be transmitted for TS1/TS2 Ordered
    Sets.
    """
    def __init__(self, *, set_data, first_word_ctrl=0b1111, transmit_burst_length=1, include_config=False):
        self._set_data          = set_data
        self._first_word_ctrl   = first_word_ctrl
        self._total_to_transmit = transmit_burst_length
        self._include_config    = include_config

        #
        # I/O port
        #
        self.source            = USBRawSuperSpeedStream()

        self.start             = Signal()
        self.done              = Signal()

        if self._include_config:
            self.request_hot_reset      = Signal()
            self.request_loopback       = Signal()
            self.request_no_scrambling  = Signal()



    def elaborate(self, platform):
        m = Module()

        # Keep track of how many ordered sets we've sent.
        sent_ordered_sets = Signal(range(self._total_to_transmit))

        with m.FSM(domain="ss"):

            # IDLE - we're currently waiting for a ``start`` request.=
            with m.State("IDLE"):

                # Once we get our start request, start blasting out our words in order.
                with m.If(self.start):
                    m.next = "WORD_0"

            # Sequentially output each word of our training set.
            for i in range(len(self._set_data)):
                with m.State(f"WORD_{i}"):
                    is_last_word = (i + 1 == len(self._set_data))
                    m.d.comb += [
                        self.source.valid  .eq(1),
                        self.source.data   .eq(self._set_data[i]),
                        self.source.ctrl   .eq(self._first_word_ctrl if i == 0 else 0b0000),
                        self.source.first  .eq(i == 0),
                        self.source.last   .eq(is_last_word)
                    ]

                    # If this is the first data word of our Training Set, we'll optionally set
                    # our control fields.
                    if self._include_config and (i == 1):
                        with m.If(self.request_hot_reset):
                            m.d.comb += self.source.data.word_select(1, 8)[0].eq(1)
                        with m.If(self.request_loopback):
                            m.d.comb += self.source.data.word_select(1, 8)[2].eq(1)
                        with m.If(self.request_no_scrambling):
                            m.d.comb += self.source.data.word_select(1, 8)[3].eq(1)


                    with m.If(self.source.ready):
                        # If we're generating the state for the last word,
                        # we also have to handle our count.
                        if is_last_word:
                            # If we've just reached the total number of sets, we're done!
                            # Reset our count to zero, and indicate done-ness.
                            with m.If(sent_ordered_sets + 1 == self._total_to_transmit):
                                m.d.comb += self.done.eq(1)
                                m.d.ss   += sent_ordered_sets.eq(0)

                                # If we're still requesting data be sent, restart from the first word.
                                with m.If(self.start):
                                    m.next = "WORD_0"
                                # Otherwise, return to idle until we receive ``start`` again`.
                                with m.Else():
                                    m.next = "IDLE"

                            # If we're not yet done transmitting our full set, continue.
                            with m.Else():
                                m.d.ss += sent_ordered_sets.eq(sent_ordered_sets + 1)
                                m.next = "WORD_0"

                        else:
                            m.next = f"WORD_{i+1}"

        return m


class TSTransceiver(Elaboratable):
    """Training Sequence Unit

    Detect/generate the Training Sequence Ordered Sets required for a USB3.0 link with simple
    control/status signals.
    """
    def __init__(self):

        #
        # I/O port
        #
        self.sink       = USBRawSuperSpeedStream()
        self.source     = USBRawSuperSpeedStream()

        # Detectors
        self.tseq_detected         = Signal() # o
        self.ts1_detected          = Signal() # o
        self.inverted_ts1_detected = Signal() # o
        self.ts2_detected          = Signal() # o

        self.hot_reset_requested   = Signal()
        self.loopback_requested    = Signal()
        self.no_scrambling_requested = Signal()

        # Emitters
        self.send_tseq_burst       = Signal() # i
        self.send_ts1_burst        = Signal() # i
        self.send_ts2_burst        = Signal() # i
        self.transmitting          = Signal() # o
        self.burst_complete        = Signal() # o

        self.request_hot_reset     = Signal()
        self.request_loopback      = Signal()
        self.request_no_scrambling = Signal()


    def elaborate(self, platform):
        m = Module()

        #
        # TSEQ Detectors
        #

        # We're always ready to accept data.
        m.d.comb += self.sink.ready.eq(1)

        # TSEQ detector (not strictly required, but useful for alignment)
        m.submodules.tseq_detector = tseq_detector = TSBurstDetector(
            set_data        = TSEQ_SET_DATA,
            first_word_ctrl = 0b0001,
            sets_in_burst   = 32
        )
        m.d.comb += [
            tseq_detector.sink  .tap(self.sink),
            self.tseq_detected  .eq(tseq_detector.detected)
        ]

        # TS1 detector
        m.submodules.ts1_detector = ts1_detector = TSBurstDetector(
            set_data        = TS1_SET_DATA,
            sets_in_burst   = 8
        )
        m.d.comb += [
            ts1_detector.sink   .tap(self.sink),
            self.ts1_detected   .eq(ts1_detector.detected)
        ]

        # Inverted TS1 detector
        m.submodules.inverted_ts1_detector = inverted_ts1_detector = TSBurstDetector(
            set_data        = INVERTED_TS1_SET_DATA,
            sets_in_burst   = 8
        )
        m.d.comb += [
            inverted_ts1_detector.sink  .tap(self.sink),
            self.inverted_ts1_detected  .eq(inverted_ts1_detector.detected)
        ]

        # TS2 detector
        m.submodules.ts2_detector = ts2_detector = TSBurstDetector(
            set_data        = TS2_SET_DATA,
            sets_in_burst   = 8,
            include_config  = True
        )
        m.d.comb += [
            ts2_detector.sink   .tap(self.sink),
            self.ts2_detected   .eq(ts2_detector.detected),

            self.hot_reset_requested    .eq(ts2_detector.hot_reset),
            self.loopback_requested     .eq(ts2_detector.loopback_requested),
            self.no_scrambling_requested.eq(ts2_detector.scrambling_disabled),
        ]


        #
        # Ordered set generators
        #

        # TSEQ generator
        m.submodules.tseq_generator = tseq_generator = TSEmitter(
            set_data              = TSEQ_SET_DATA,
            first_word_ctrl       = 0b0001,
            transmit_burst_length = 65536
        )
        with m.If(self.send_tseq_burst):
            m.d.comb += [
                tseq_generator.start .eq(1),
                self.source          .stream_eq(tseq_generator.source)
            ]

        # TS1 generator
        m.submodules.ts1_generator = ts1_generator = TSEmitter(
            set_data              = TS1_SET_DATA,
            transmit_burst_length = 16
        )
        with m.If(self.send_ts1_burst):
            m.d.comb += [
                ts1_generator.start  .eq(1),
                self.source          .stream_eq(ts1_generator.source)
            ]

        # TS2 Generator
        m.submodules.ts2_generator = ts2_generator = TSEmitter(
            set_data              = TS2_SET_DATA,
            transmit_burst_length = 16,
            include_config        = True,
        )
        with m.If(self.send_ts2_burst):
            m.d.comb += [
                ts2_generator.start                 .eq(1),
                ts2_generator.request_hot_reset     .eq(self.request_hot_reset),
                ts2_generator.request_loopback      .eq(self.request_loopback),
                ts2_generator.request_no_scrambling .eq(self.request_no_scrambling),
                self.source                         .stream_eq(ts2_generator.source),
            ]

        #
        # Status signals.
        #
        tseq_transmitting = tseq_generator.source.valid
        ts1_transmitting  = ts1_generator.source.valid
        ts2_transmitting  = ts2_generator.source.valid
        m.d.comb += [
            self.transmitting   .eq(tseq_transmitting   | ts1_transmitting   | ts2_transmitting),
            self.burst_complete .eq(tseq_generator.done | ts1_generator.done | ts2_generator.done),
        ]

        return m
